# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
from tqdm import tqdm

from hyperspy.signals import Signal2D

from pyxem.signals import DiffractionVectors, ElectronDiffraction2D
from pyxem.utils._signals import _transfer_signal_axes
from pyxem.utils.segment_utils import (
    norm_cross_corr,
    separate_watershed,
    get_gaussian2d,
)


class LearningSegment:
    """Signal classes for nanocrystal segmentation results obtained from
    i) machine learning (LearningSegment) and ii) VDF images (VDFSegment).
    """

    _signal_type = "learning_segment"

    def __init__(self, factors, loadings):
        # Factors as Signal2D
        self.factors = factors
        # Corresponding loadings as Signal2D
        self.loadings = loadings

    def get_ncc_matrix(self):
        """Get the normalised correlation coefficient (NCC) matrix containing
        the NCC between each pair of segments.

        Returns
        -------
        ncc_matrix : Signal2D
            Normalised correlation coefficient matrix for loadings and factors.
        """
        # Set up empty matrices of correct size to store NCC values.
        num_comp = np.shape(self.loadings.data)[0]
        ncc_loadings = np.zeros((num_comp, num_comp))
        ncc_factors = np.zeros((num_comp, num_comp))
        factors = self.factors.map(np.nan_to_num, inplace=False).copy()
        loadings = self.loadings.map(np.nan_to_num, inplace=False).copy()
        # Iterate through loadings calculating NCC values.
        for i in np.arange(num_comp):
            ncc_loadings[i] = list(
                map(
                    lambda x: norm_cross_corr(x, template=loadings.data[i]),
                    loadings.data,
                )
            )
        # Iterate through factors calculating NCC values.
        for i in np.arange(num_comp):
            ncc_factors[i] = list(
                map(
                    lambda x: norm_cross_corr(x, template=factors.data[i]), factors.data
                )
            )
        # Convert matrix to Signal2D and set axes
        ncc_sig = Signal2D(np.array((ncc_loadings, ncc_factors)))
        ncc_sig.axes_manager.signal_axes[0].name = "index"
        ncc_sig.axes_manager.signal_axes[1].name = "index"
        ncc_sig.metadata.General.title = "Normalised Correlation Coefficient"

        return ncc_sig

    def correlate_learning_segments(self, corr_th_factors=0.4, corr_th_loadings=0.4):
        """Iterates through the factors and loadings and calculates the
        normalized cross-correlation between all factors and all
        loadings. Factors and loadings are summed if the correlations of
        both factors and loadings exceed the given thresholds.

        Parameters
        ----------
        corr_th_factors : float
            Correlation threshold for factors. Must be between -1 and 1.
            Factors and loadings are summed if both factors and loadings
            have normalized cross-correlations above corr_th_factors and
            corr_th_loadings, respectively.
        corr_th_loadings : int, optional
            Correlation threshold for loadings. Must be between -1 and 1.

        Returns
        -------
        learning_segment : LearningSegment
            LearningSegment where possibly some factors and loadings
            have been summed.
        """
        # If a mask was used during the decomposition, the factors and/or
        # loadings will contain nan, which must be converted to numbers prior
        # to the correlations calculations.
        factors = self.factors.map(np.nan_to_num, inplace=False)
        loadings = self.loadings.map(np.nan_to_num, inplace=False)
        factors = factors.copy().data
        loadings = loadings.copy().data
        correlated_loadings = np.zeros_like(loadings[:1])
        correlated_factors = np.zeros_like(factors[:1])

        # For each loading and factor, calculate the normalized
        # cross-correlation to all other loadings and factors, and define
        # add_indices for those with a value above corr_th_loadings and
        # corr_th_factors respectively.
        while np.shape(loadings)[0] > 0:
            corr_list_loadings = list(
                map(lambda x: norm_cross_corr(x, template=loadings[0]), loadings)
            )
            corr_list_factors = list(
                map(lambda x: norm_cross_corr(x, template=factors[0]), factors)
            )

            add_indices = np.where(
                list(
                    map(
                        lambda l, f: (l > corr_th_loadings and f > corr_th_factors),
                        corr_list_loadings,
                        corr_list_factors,
                    )
                )
            )

            correlated_loadings = np.append(
                correlated_loadings,
                np.array([np.sum(loadings[add_indices], axis=0)]),
                axis=0,
            )
            correlated_factors = np.append(
                correlated_factors,
                np.array([np.sum(factors[add_indices], axis=0)]),
                axis=0,
            )

            loadings = np.delete(loadings, add_indices, axis=0)
            factors = np.delete(factors, add_indices, axis=0)

        correlated_loadings = Signal2D(np.delete(correlated_loadings, 0, axis=0))
        correlated_factors = Signal2D(np.delete(correlated_factors, 0, axis=0))
        learning_segment = LearningSegment(
            factors=correlated_factors, loadings=correlated_loadings
        )
        return learning_segment

    def separate_learning_segments(
        self,
        min_intensity_threshold=0,
        min_distance=2,
        min_size=10,
        max_size=np.inf,
        max_number_of_grains=np.inf,
        marker_radius=2,
        threshold=False,
        exclude_border=False,
    ):
        """Segmentation of loading maps by the watershed
        segmentation method implemented in scikit-image [1,2].

        Parameters
        ----------
        min_intensity_threshold : float
            Loading segments with a maximum intensity below
            min_intensity_threshold are discarded.
        min_distance: int
            Minimum distance (in pixels) between grains required for
            them to be considered as separate grains.
        min_size : float
            Grains with size (i.e. total number of pixels) below
            min_size are discarded.
        max_size : float
            Grains with size (i.e. total number of pixels) above
            max_size are discarded.
        max_number_of_grains : int
            Maximum number of grains included in the returned separated
            grains. If it is exceeded, those with highest peak
            intensities will be returned.
        marker_radius : float
            If 1 or larger, each marker for watershed is expanded to a disk
            of radius marker_radius. marker_radius should not exceed
            2*min_distance.
        threshold: bool
            If True, a mask is calculated by thresholding the VDF image
            by the Li threshold method in scikit-image. If False
            (default), the mask is the boolean VDF image.
        exclude_border : int or True, optional
            If non-zero integer, peaks within a distance of
            exclude_border from the boarder will be discarded. If True,
            peaks at or closer than min_distance of the boarder, will be
            discarded.

        References
        ----------
        [1] http://scikit-image.org/docs/dev/auto_examples/segmentation/plot_watershed.html
        [2] https://scikit-image.org/docs/dev/auto_examples/segmentation/plot_expand_labels.html#sphx-glr-auto-examples-segmentation-plot-expand-labels-py

        Returns
        -------
        learning_segment : LearningSegment
            LearningSegment where the loadings have been segmented and some
            factors have been repeated according to the new number of loading
            segments.
        """

        factors = self.factors.copy()
        loadings = self.loadings.copy()
        loadings_shape_x = loadings.data.shape[1]
        loadings_shape_y = loadings.data.shape[2]
        factors_shape_x = factors.data.shape[1]
        factors_shape_y = factors.data.shape[2]

        loadings_segments = np.array(
            loadings.map(
                separate_watershed,
                show_progressbar=True,
                inplace=False,
                min_distance=min_distance,
                min_size=min_size,
                max_size=max_size,
                max_number_of_grains=max_number_of_grains,
                marker_radius=marker_radius,
                threshold=threshold,
                exclude_border=exclude_border,
                ragged=True,
            ),
            dtype=object,
        )

        segments, factors_of_segments = [], []
        num_segs_tot = 0
        for loading_segment, factor in zip(loadings_segments, factors):
            segments = np.append(segments, loading_segment)
            num_segs = np.shape(loading_segment)[0]
            factors_of_segments = np.append(
                factors_of_segments,
                np.broadcast_to(factor, (num_segs, factors_shape_x, factors_shape_y)),
            )
            num_segs_tot += num_segs
        segments = segments.reshape((num_segs_tot, loadings_shape_y, loadings_shape_x))
        factors_of_segments = factors_of_segments.reshape(
            (num_segs_tot, factors_shape_x, factors_shape_y)
        )

        delete_indices = list(
            map(lambda x: x.max() < min_intensity_threshold, segments)
        )
        delete_indices = np.where(delete_indices)
        segments = np.delete(segments, delete_indices, axis=0)
        factors_of_segments = np.delete(factors_of_segments, delete_indices, axis=0)

        # if TraitError is raised, it is likely no segements were found
        segments = Signal2D(segments).transpose(navigation_axes=[0], signal_axes=[2, 1])
        factors_of_segments = Signal2D(factors_of_segments)
        learning_segment = LearningSegment(
            factors=factors_of_segments, loadings=segments
        )
        return learning_segment


class VDFSegment:
    """Signal classes for nanocrystal segmentation results obtained from
    VDF images (VDFSegment).
    """

    _signal_type = "vdf_segment"

    def __init__(self, segments, vectors_of_segments, intensities=None):
        # Segments as Signal2D
        self.segments = segments
        # DiffractionVectors
        self.vectors_of_segments = vectors_of_segments
        # Intensities corresponding to each vector
        self.intensities = intensities

    def get_ncc_matrix(self):
        """Get the normalised correlation coefficient (NCC) matrix containing
        the NCC between each pair of segments.

        Returns
        -------
        ncc_matrix : Signal2D
            Normalised correlation coefficient matrix.
        """
        # TODO: This code should be factored out for reuse in other ncc method.
        # Set up an empty matrix of correct size to store NCC values.
        num_comp = np.shape(self.segments.data)[0]
        ncc_matrix = np.zeros((num_comp, num_comp))
        # Iterate through segments calculating NCC values.
        for i in np.arange(num_comp):
            ncc_matrix[i] = list(
                map(
                    lambda x: norm_cross_corr(x, template=self.segments.data[i]),
                    self.segments.data,
                )
            )
        # Convert matrix to Signal2D and set axes
        ncc_sig = Signal2D(ncc_matrix)
        ncc_sig.axes_manager.signal_axes[0].name = "segment index"
        ncc_sig.axes_manager.signal_axes[1].name = "segment index"
        ncc_sig.metadata.General.title = "Normalised Correlation Coefficient"
        return ncc_sig

    def correlate_vdf_segments(
        self, corr_threshold=0.7, vector_threshold=4, segment_threshold=3
    ):
        """Iterates through VDF segments and sums those that are
        associated with the same segment. Summation will be done for
        those segments that have a normalised cross correlation above
        corr_threshold. The vectors of each segment sum will be updated
        accordingly, so that the vectors of each resulting segment sum
        are all the vectors of the original individual segments. Each
        vector is assigned an intensity that is the integrated intensity
        of the segment it originates from.

        Parameters
        ----------
        corr_threshold : float
            Segments will be summed if they have a normalized cross-
            correlation above corr_threshold. Must be between 0 and 1.
        vector_threshold : int, optional
            Correlated segments having a number of vectors less than
            vector_threshold will be discarded.
        segment_threshold : int, optional
            Correlated segment intensities that lie in a region where
            a number of segments less than segment_thresholdhave been
            found, are set to 0, i.e. the resulting segment will only
            have intensities above 0 where at least a number of
            segment_threshold segments have intensitives above 0.

        Returns
        -------
        vdfseg : VDFSegment
            The VDFSegment instance updated according to the image
            correlation results.
        """
        vectors = self.vectors_of_segments.data

        if segment_threshold > vector_threshold:
            raise ValueError(
                "segment_threshold must be smaller than or "
                "equal to vector_threshold."
            )

        segments = self.segments.data.copy()
        num_vectors = np.shape(vectors)[0]
        gvectors = np.array(np.empty(num_vectors, dtype=object))
        vector_indices = np.array(np.empty(num_vectors, dtype=object))

        for i in np.arange(num_vectors):
            gvectors[i] = np.array(vectors[i].copy())
            vector_indices[i] = np.array([i], dtype=int)

        correlated_segments = np.zeros_like(segments[:1])
        correlated_vectors = np.array([0.0], dtype=object)
        correlated_vectors[0] = np.array(np.zeros_like(vectors[:1]))
        correlated_vector_indices = np.array([0], dtype=object)
        correlated_vector_indices[0] = np.array([0])
        i = 0
        pbar = tqdm(total=np.shape(segments)[0])
        while np.shape(segments)[0] > i:
            # For each segment, calculate the normalized cross-correlation to
            # all other segments, and define add_indices for those with a value
            # above corr_threshold.
            corr_list = list(
                map(lambda x: norm_cross_corr(x, template=segments[i]), segments)
            )

            corr_add = list(map(lambda x: x > corr_threshold, corr_list))
            add_indices = np.where(corr_add)
            # If there are more add_indices than vector_threshold,
            # sum segments and add their vectors. Otherwise, discard segment.
            if (
                np.shape(add_indices[0])[0] >= vector_threshold
                and np.shape(add_indices[0])[0] > 1
            ):
                new_segment = np.array([np.sum(segments[add_indices], axis=0)])
                if segment_threshold > 1:
                    segment_check = np.zeros_like(segments[add_indices], dtype=int)
                    segment_check[np.where(segments[add_indices])] = 1
                    segment_check = np.sum(segment_check, axis=0, dtype=int)
                    segment_mask = np.zeros_like(segments[0], dtype=bool)
                    segment_mask[np.where(segment_check >= segment_threshold)] = 1
                    new_segment = new_segment * segment_mask
                correlated_segments = np.append(
                    correlated_segments, new_segment, axis=0
                )
                add_indices = add_indices[0]
                new_vectors = np.array([0], dtype=object)
                new_vectors[0] = np.concatenate(gvectors[add_indices], axis=0).reshape(
                    -1, 2
                )
                correlated_vectors = np.append(correlated_vectors, new_vectors, axis=0)
                new_indices = np.array([0], dtype=object)
                new_indices[0] = np.concatenate(
                    vector_indices[add_indices], axis=0
                ).reshape(-1, 1)
                correlated_vector_indices = np.append(
                    correlated_vector_indices, new_indices, axis=0
                )
            elif np.shape(add_indices[0])[0] >= vector_threshold:
                add_indices = add_indices[0]
                correlated_segments = np.append(
                    correlated_segments, segments[add_indices], axis=0
                )
                correlated_vectors = np.append(
                    correlated_vectors, gvectors[add_indices], axis=0
                )
                correlated_vector_indices = np.append(
                    correlated_vector_indices, vector_indices[add_indices], axis=0
                )
            else:
                add_indices = i
            segments = np.delete(segments, add_indices, axis=0)
            gvectors = np.delete(gvectors, add_indices, axis=0)
            vector_indices = np.delete(vector_indices, add_indices, axis=0)

        pbar.close()
        correlated_segments = np.delete(correlated_segments, 0, axis=0)
        correlated_vectors = np.delete(correlated_vectors, 0, axis=0)
        correlated_vector_indices = np.delete(correlated_vector_indices, 0, axis=0)
        correlated_vector_intensities = np.array(
            np.empty(len(correlated_vectors)), dtype=object
        )

        # Sum the intensities in the original segments and assign those to the
        # correct vectors by referring to vector_indices.
        # If segment_mask has been used, use the segments as masks too.
        if segment_threshold > 1:
            for i in range(len(correlated_vectors)):
                correlated_vector_intensities[i] = np.zeros(
                    len(correlated_vector_indices[i])
                )
                segment_mask = np.zeros_like(segment_mask)
                segment_mask[np.where(correlated_segments[i])] = 1
                segment_intensities = np.sum(
                    self.segments.data * segment_mask, axis=(1, 2)
                )
                for n, index in zip(
                    range(len(correlated_vector_indices[i])),
                    correlated_vector_indices[i],
                ):
                    correlated_vector_intensities[i][n] = np.sum(
                        segment_intensities[index]
                    )
        else:
            segment_intensities = np.sum(self.segments.data, axis=(1, 2))
            for i in range(len(correlated_vectors)):
                correlated_vector_intensities[i] = np.zeros(
                    len(correlated_vector_indices[i])
                )
                for n, index in zip(
                    range(len(correlated_vector_indices[i])),
                    correlated_vector_indices[i],
                ):
                    correlated_vector_intensities[i][n] = np.sum(
                        segment_intensities[index]
                    )

        vdfseg = VDFSegment(
            Signal2D(correlated_segments),
            DiffractionVectors(correlated_vectors),
            correlated_vector_intensities,
        )

        # Transfer axes properties of segments
        vdfseg.segments = _transfer_signal_axes(vdfseg.segments, self.segments)
        n = vdfseg.segments.axes_manager.navigation_axes[0]
        n.name = "n"
        n.units = "number"

        return vdfseg

    def get_virtual_electron_diffraction(self, calibration, shape, sigma):
        """Obtain a virtual electron diffraction signal that consists
        of one virtual diffraction pattern for each segment. The virtual
        diffraction pattern is composed of Gaussians centered at each
        vector position. If given, the integrated intensities of each
        vector will be taken into account by multiplication with the
        Gaussians.

        Parameters
        ----------
        calibration : float
            Reciprocal space calibration in inverse Angstrom per pixel.
        shape : tuple
            Shape of the signal, (shape_x, shape_y) in pixels, where
            shape_x and shape_y are integers.
        sigma : float
            The standard deviation of the Gaussians in inverse Angstrom
            per pixel. 'calibration' is a decent starting value.

        Returns
        -------
        virtual_ed : ElectronDiffraction2D
            Virtual electron diffraction signal consisting of one
            virtual diffraction pattern for each segment.
        """
        vectors = self.vectors_of_segments.data
        segments = self.segments.data
        num_segments = np.shape(segments)[0]

        if self.intensities is None:
            raise ValueError(
                "The VDFSegment does not have the attribute "
                "intensities, required for this method."
            )
        else:
            intensities = self.intensities

        # TODO: Refactor this to use the diffsims simulation to plot functionality
        size_x, size_y = shape[0], shape[1]
        cx, cy = -size_x / 2 * calibration, -size_y / 2 * calibration
        x, y = np.indices((size_x, size_y))
        x, y = x * calibration + cx, y * calibration + cy
        virtual_ed = np.zeros((size_x, size_y, num_segments))

        for i in range(num_segments):
            # Allow plotting for segments that are only associated with
            # one vector.
            if np.shape(np.shape(vectors[i]))[0] <= 1:
                virtual_ed[..., i] = get_gaussian2d(
                    intensities[i],
                    vectors[i][..., 0],
                    vectors[i][..., 1],
                    x=x,
                    y=y,
                    sigma=sigma,
                )
            # Allow plotting for segments associated with several vectors.
            else:
                virtual_ed[..., i] = sum(
                    list(
                        map(
                            lambda a, xo, yo: get_gaussian2d(
                                a, xo, yo, x=x, y=y, sigma=sigma
                            ),
                            intensities[i],
                            vectors[i][..., 0],
                            vectors[i][..., 1],
                        )
                    )
                )

        virtual_ed = ElectronDiffraction2D(virtual_ed.T)

        return virtual_ed



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Helper functions and classes for managing pyxem.
This module and documentation is only relevant for pyxem developers,
not for users.
.. warning:
    This module and its submodules are for internal use only.  Do not
    use them in your own code. We may change the API at any time with no
    warning.
"""

import functools
import inspect
from typing import Callable, Optional, Union
import warnings

from pyxem.common import VisibleDeprecationWarning


class deprecated:
    """Decorator to mark deprecated functions with an informative
    warning.
    Adapted from
    `scikit-image
    <https://github.com/scikit-image/scikit-image/blob/main/skimage/_shared/utils.py>`_
    and `matplotlib
    <https://github.com/matplotlib/matplotlib/blob/master/lib/matplotlib/_api/deprecation.py>`_.
    """

    def __init__(
        self,
        since: Union[str, int, float],
        alternative: Optional[str] = None,
        alternative_is_function: bool = True,
        removal: Union[str, int, float, None] = None,
    ):
        """Visible deprecation warning.
        Parameters
        ----------
        since
            The release at which this API became deprecated.
        alternative
            An alternative API that the user may use in place of the
            deprecated API.
        alternative_is_function
            Whether the alternative is a function. Default is ``True``.
        removal
            The expected removal version.
        """
        self.since = since
        self.alternative = alternative
        self.alternative_is_function = alternative_is_function
        self.removal = removal

    def __call__(self, func: Callable):
        # Wrap function to raise warning when called, and add warning to
        # docstring
        if self.alternative is not None:
            if self.alternative_is_function:
                alt_msg = f" Use `{self.alternative}()` instead."
            else:
                alt_msg = f" Use `{self.alternative}` instead."
        else:
            alt_msg = ""
        if self.removal is not None:
            rm_msg = f" and will be removed in version {self.removal}"
        else:
            rm_msg = ""
        msg = f"Function `{func.__name__}()` is deprecated{rm_msg}.{alt_msg}"

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            warnings.simplefilter(
                action="always", category=VisibleDeprecationWarning, append=True
            )
            func_code = func.__code__
            warnings.warn_explicit(
                message=msg,
                category=VisibleDeprecationWarning,
                filename=func_code.co_filename,
                lineno=func_code.co_firstlineno + 1,
            )
            return func(*args, **kwargs)

        # Modify docstring to display deprecation warning
        old_doc = inspect.cleandoc(func.__doc__ or "").strip("\n")
        notes_header = "\nNotes\n-----"
        new_doc = (
            f"[*Deprecated*] {old_doc}\n"
            f"{notes_header if notes_header not in old_doc else ''}\n"
            f".. deprecated:: {self.since}\n"
            f"   {msg.strip()}"  # Matplotlib uses three spaces
        )
        wrapped.__doc__ = new_doc

        return wrapped


class deprecated_argument:
    """Decorator to remove an argument from a function or method's
    signature.
    Adapted from `scikit-image
    <https://github.com/scikit-image/scikit-image/blob/main/skimage/_shared/utils.py>`_.
    """

    def __init__(self, name, since, removal, alternative=None):
        self.name = name
        self.since = since
        self.removal = removal
        self.alternative = alternative

    def __call__(self, func):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            if self.name in kwargs.keys():
                msg = (
                    f"Argument `{self.name}` is deprecated and will be removed in "
                    f"version {self.removal}. To avoid this warning, please do not use "
                    f"`{self.name}`. "
                )
                if self.alternative is not None:
                    msg += f"Use `{self.alternative}` instead. "
                    kwargs[self.alternative] = kwargs.pop(self.name)
                msg += f"See the documentation of `{func.__name__}()` for more details."
                warnings.simplefilter(
                    action="always", category=VisibleDeprecationWarning
                )
                func_code = func.__code__
                warnings.warn_explicit(
                    message=msg,
                    category=VisibleDeprecationWarning,
                    filename=func_code.co_filename,
                    lineno=func_code.co_firstlineno + 1,
                )
            return func(*args, **kwargs)

        return wrapped



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Generating subpixel resolution on diffraction vectors."""

import numpy as np
from skimage import morphology
from skimage.measure import label
from scipy import ndimage as ndi
from scipy.ndimage import center_of_mass

from hyperspy.signals import BaseSignal

import logging

_logger = logging.getLogger(__name__)

from pyxem.generators.subpixelrefinement_generator import _get_pixel_vectors
from pyxem.signals import DiffractionVectors


def _get_intensities(z, vectors, radius=1):
    """Basic intensity integration routine, takes the maximum value at the
    given vector positions with the number of pixels given by `radius`.

    Parameters
    ----------
    vectors : DiffractionVectors
        Vectors to the locations of the spots to be
        integrated.
    radius: int,
        Number of pixels within which to find the largest maximum

    Returns
    -------
    intensities : np.array
        List of extracted intensities
    """
    i, j = np.array(vectors.data).astype(int).T

    if radius > 1:
        footprint = morphology.disk(radius)
        filtered = ndi.maximum_filter(z, footprint=footprint)
        intensities = filtered[j, i].reshape(-1, 1)  # note that the indices are flipped
    else:
        intensities = z[j, i].reshape(-1, 1)  # note that the indices are flipped

    return np.array(intensities)


def _take_ragged(z, indices, _axis=None, out=None, mode="raise"):
    """Like `np.take` for ragged arrays, see `np.take` for documentation."""
    return np.take(z, indices, axis=_axis, out=out, mode=mode)


def _get_largest_connected_region(segmentation):
    """Take a binary segmentation image and return the largest connected area."""
    labels = label(segmentation)
    largest = np.argmax(np.bincount(labels.flat, weights=segmentation.flat))
    return (labels == largest).astype(int)


def _get_intensities_summation_method(
    z,
    vectors,
    box_inner: int = 7,
    box_outer: int = 10,
    n_min: int = 5,
    n_max: int = None,
    snr_thresh=3.0,
    verbose: bool = False,
):
    """Integrate reflections using the summation method.

    Two boxes are defined, the inner box is used to define the
    integration area. The outer box is used to calculate the
    average signal-to-noise ratio (SNR).

    All pixels with a large enough SNR are considered to be signal. The largest region
    of connected signal pixels are summed to calculate the reflection intensity.

    Parameters
    ----------
    vectors : DiffractionVectors
        Vectors to the locations of the spots to be
        integrated.
    box_inner : int
        Defines the radius (size) of the inner box, which must be larger than the reflection.
        The total box size is 2*box_inner
    box_outer : int
        Defines radius (size) of the outer box. The total box size is 2*box_inner
        The border between the inner and outer box is considered background
        and used to calculate the (SNR) for each pixel: SNR = (I - <I>/std(I_bkg)).
    snr_thresh : float
        Minimum signal-to-noise for a pixel to be considered as `signal`.
    n_min: int
        If the number of SNR pixels in the inner box < n_min, the reflection is discared
    n_max:
        If the number of SNR pixels in the inner box >= n_max, the reflection is discareded
        Defaults to the inner box size (`box_inner**2`.
    verbose : bool
        Print statistics for every reflection (for debugging)

    Returns
    -------
    peaks : np.array
        Array with 4 columns: X-position, Y-position, intensity, reflection SNR

    Notes
    -----
    Implementation based on Barty et al, J. Appl. Cryst. (2014). 47, 1118-1131
    Lesli, Acta Cryst. (2006). D62, 48-57

    """
    if not n_max:  # pragma: no cover
        n_max = box_inner**2

    peaks = []

    for i, j in vectors:
        box = z[j - box_inner : j + box_inner, i - box_inner : i + box_inner].copy()

        bkg = np.hstack(
            [
                z[j - box_outer : j + box_outer, i - box_outer : i - box_inner].ravel(),
                z[j - box_outer : j + box_outer, i + box_inner : i + box_outer].ravel(),
                z[j - box_outer : j - box_inner, i - box_inner : i + box_inner].ravel(),
                z[j + box_inner : j + box_outer, i - box_inner : i + box_inner].ravel(),
            ]
        )

        bkg_mean = bkg.mean()
        bkg_std = bkg.std()
        box_snr = (box - bkg_mean) / bkg_std

        # get mask for signal (I > SNR)
        signal_mask = _get_largest_connected_region(box_snr > snr_thresh)

        n_pix = signal_mask.sum()
        signal = (box - bkg_mean) * signal_mask
        inty = signal.sum()
        snr = (inty / n_pix) / bkg_std
        sigma = inty / snr

        # calculate center of mass
        com_X, com_Y = center_of_mass(box, labels=signal_mask, index=1)
        dX = com_X - box_inner
        dY = com_Y - box_inner

        X = i + dX
        Y = j + dY

        if verbose:  # pragma: no cover
            print(
                f"\nMean(I): {bkg_mean:.2f} | Std(I): {bkg_std:.2f} | n_pix: {n_pix} \n"
                f"I: {inty:.2f} | Sigma(I): {sigma:.2f} | SNR(I): {snr:.2f} | I/pix: {inty/n_pix:.2f} \n"
                f"i: {i:.2f} | j: {j:.2f} | dX: {dX:.2f} | dY: {dY:.2f} | X: {X:.2f} | Y: {Y:.2f} "
            )
            # for debugging purposes
            import matploltib.pyplot as plt

            plt.imshow(signal)
            plt.plot(dY + box_inner, dX + box_inner, "r+")  # center_of_mass
            plt.plot(box_inner, box_inner, "g+")  # input
            plt.show()

        if n_pix > n_max:  # pragma: no cover
            continue
        if n_pix < n_min:  # pragma: no cover
            continue

        # for some reason X/Y are reversed here
        peaks.append([Y, X, inty, sigma])

    peaks = np.array(peaks)

    return np.array(peaks)


class IntegrationGenerator:
    """Integrates reflections at the given vector positions.

    Parameters
    ----------
    dp : ElectronDiffraction2D
        The electron diffraction patterns to be refined
    vectors : DiffractionVectors | ndarray
        Vectors (in calibrated units) to the locations of the spots to be
        integrated. If given as DiffractionVectors, it must have the same
        navigation shape as the electron diffraction patterns. If an ndarray,
        the same set of vectors is mapped over all electron diffraction
        patterns.
    """

    def __init__(self, dp, vectors):
        self.dp = dp
        self.vectors_init = vectors
        self.last_method = None
        sig_ax = dp.axes_manager.signal_axes
        self.calibration = [sig_ax[0].scale, sig_ax[1].scale]
        self.center = [sig_ax[0].size / 2, sig_ax[1].size / 2]

        self.vector_pixels = _get_pixel_vectors(
            dp=dp, vectors=vectors, calibration=self.calibration, center=self.center
        )

    def extract_intensities(self, radius: int = 1):
        """Basic intensity integration routine, takes the maximum value at the
        given vector positions with the number of pixels given by `radius`.

        Parameters
        ----------
        radius: int,
            Number of pixels within which to find the largest maximum

        Returns
        -------
        intensities : :obj:`hyperspy.signals.BaseSignal`
            List of extracted intensities
        """
        intensities = self.dp.map(
            _get_intensities, vectors=self.vector_pixels, radius=radius, inplace=False
        )

        intensities = BaseSignal(intensities)
        intensities = intensities.transpose(0)

        return intensities

    def extract_intensities_summation_method(
        self,
        box_inner: int = 7,
        box_outer: int = 10,
        n_min: int = 5,
        n_max: int = 1000,
        snr_thresh: float = 3.0,
    ):
        """Integrate reflections using the summation method. Two boxes are defined,
        the inner box is used to define the integration area. The outer box is used
        to calculate the average signal-to-noise ratio (SNR).
        All pixels with a large enough SNR are considered to be signal. The largest region
        of connected signal pixels are summed to calculate the reflection intensity. The
        diffraction vectors are calculated as the center of mass of the signal pixels.

        Parameters
        ----------
        box_inner : int
            Defines the size of the inner box, which must be larger than the reflection.
        box_outer : int
            Defines the size of the outer box. The border between the inner and outer
            box is considered background and used to calculate the (SNR) for each
            pixel: SNR = (I - <I>/std(I_bkg)).
        snr_thresh : float
            Minimum signal-to-noise for a pixel to be considered as `signal`.
        n_min: int
            If the number of SNR pixels in the inner box < n_min, the reflection is discared
        n_max:
            If the number of SNR pixels in the inner box > n_max, the reflection is discareded
        verbose : bool
            Print statistics for every reflection (for debugging)

        Returns
        -------
        vectors : :obj:`pyxem.signals.diffraction_vectors.DiffractionVectors`
            DiffractionVectors with optimized coordinates, where the attributes
            vectors.intensities -> `I`, vectors.sigma -> `sigma(I)`, and
            vectors.snr -> `I / sigma(I)`

        Notes
        -----
        Implementation based on Barty et al, J. Appl. Cryst. (2014). 47, 1118-1131
                                Lesli, Acta Cryst. (2006). D62, 48-57
        """
        _logger.warning(
            "This function might not work properly at the moment, check that the "
            "returned results looks reasonable."
        )
        result = self.dp.map(
            _get_intensities_summation_method,
            vectors=self.vector_pixels,
            box_inner=box_inner,
            box_outer=box_outer,
            n_min=n_min,
            n_max=n_max,
            snr_thresh=snr_thresh,
            inplace=False,
            ragged=True,
        )

        peaks = result.map(
            _take_ragged, indices=[0, 1], _axis=1, inplace=False, ragged=True
        )
        intensities = result.map(
            _take_ragged, indices=2, _axis=1, inplace=False, ragged=True
        )
        sigma = result.map(_take_ragged, indices=3, _axis=1, inplace=False, ragged=True)

        vectors = DiffractionVectors.from_peaks(
            peaks, calibration=self.calibration, center=self.center
        )
        vectors.intensities = intensities
        vectors.sigma = sigma
        vectors.snr = intensities / sigma

        return vectors



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

from hyperspy.component import Component


class ReducedIntensityCorrectionComponent(Component):
    """
    A HyperSpy fitting component to fit a 4th order polynomial that goes to zero
    intensity at zero scattering to the reduced intensity profile. This is done to
    reduce the effects of multiple scattering in the pdf.

    Parameters
    ----------
    a : float
    b : float
    c : float
    d : float
        a, b, c, and d are the coefficients of the 1st, 2nd, 3rd, and 4th
        order terms respectively of the returned polynomial.

    Returns
    -------
    p : polynomial of the form ax + bx^2 + cx^3 + dx^4

    """

    def __init__(self, a=0.0, b=0.0, c=0.0, d=0.0):
        Component.__init__(self, ("a", "b", "c", "d"))

    def function(self, x):
        """
        The actual fitted function. Based on the HyperSpy Component class.
        The function "function" is called innately by HyperSpy's fitting
        algorithm, and represents the appropriate atomic scattering factor
        function presented respectively in Lobato & Van Dyck (2014).

        Parameters
        ----------
        x : array-like
            The scattering vector magnitude

        """
        a = self.a.value
        b = self.b.value
        c = self.c.value
        d = self.d.value

        p = a * x + b * (x**2) + c * (x**3) + d * (x**4)
        return p



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import os
from tempfile import TemporaryDirectory
import pytest
import numpy
import matplotlib.pyplot as plt
import hyperspy.api as hs
import pyxem as pxm


@pytest.fixture(autouse=True)
def doctest_setup_teardown(request):
    plt.ioff()
    tmp_dir = TemporaryDirectory()
    hs.preferences.General.show_progressbar = False
    org_dir = os.getcwd()
    os.chdir(tmp_dir.name)
    yield
    os.chdir(org_dir)
    tmp_dir.cleanup()
    plt.close("all")


@pytest.fixture(autouse=True)
def add_np_am(doctest_namespace):
    doctest_namespace["np"] = numpy
    doctest_namespace["hs"] = hs
    doctest_namespace["pxm"] = pxm



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest

import numpy as np
import hyperspy.api as hs
from sklearn.cluster import DBSCAN

from pyxem.signals.electron_diffraction2d import ElectronDiffraction2D
from pyxem.signals.labeled_diffraction_vectors2d import LabeledDiffractionVectors2D


class TestLabelledDiffractionVectors:
    @pytest.fixture
    def labeled_vectors(self):
        vectors = np.reshape(np.arange(1000), (100, 10))
        labels = np.repeat(np.arange(10), 10)
        vectors = np.hstack((vectors, labels[:, np.newaxis]))
        columns = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "label"]
        return LabeledDiffractionVectors2D(data=vectors, column_names=columns)

    def test_init(self, labeled_vectors):
        assert labeled_vectors.data.shape == (100, 11)

    def test_map_vectors(self, labeled_vectors):
        def func(x):
            return np.shape(x)[0]

        result = labeled_vectors.map_vectors(func, dtype=np.float64)
        assert result.shape == (10,)
        assert np.all(result == 10)  # all labels have 10 vectors

    def test_map_vectors_with_shape(self, labeled_vectors):
        def func(x):
            return np.shape(x)

        result = labeled_vectors.map_vectors(func, dtype=np.float64, shape=(2,))
        assert result.shape == (10, 2)
        assert np.all(result == [10, 11])  # all labels have 10 vectors

    @pytest.mark.parametrize("get_polygon", [True, False])
    def test_to_markers(self, labeled_vectors, get_polygon):
        s = ElectronDiffraction2D(np.zeros((10, 10, 10, 10)))
        if get_polygon:
            points, polygons = labeled_vectors.to_markers(
                signal=s, get_polygons=get_polygon
            )
            assert isinstance(polygons, hs.plot.markers.Polygons)
        else:
            points = labeled_vectors.to_markers(signal=s, get_polygon=get_polygon)
        assert isinstance(points, hs.plot.markers.Points)

    def test_cluster_labeled_vectors(self, labeled_vectors):
        scan = DBSCAN()
        clust = labeled_vectors.cluster_labeled_vectors(method=scan)
        assert clust.data.shape == (100, 12)
        assert clust.axes_manager.signal_axes[0].size == 12
        assert len(clust.column_names) == 12
        assert len(clust.units) == 12

    def test_cluster_labeled_vectors_custom(self, labeled_vectors):
        scan = DBSCAN()

        def max(arr, columns=[0, 1]):
            return np.max(arr[:, columns], axis=0)

        kwargs = {"label_index": -1, "dtype": float, "shape": (2,)}

        clust = labeled_vectors.cluster_labeled_vectors(
            method=scan, preprocessing=max, **kwargs
        )
        assert clust.data.shape == (100, 12)
        assert clust.axes_manager.signal_axes[0].size == 12
        assert len(clust.column_names) == 12
        assert len(clust.units) == 12

    def test_cluster_labeled_vectors_fail(self, labeled_vectors):
        scan = DBSCAN()
        with pytest.raises(ValueError):
            labeled_vectors.cluster_labeled_vectors(method=scan, preprocessing="fail")

    def test_plot_cluster_labeled_vectors_error(self, labeled_vectors):
        scan = DBSCAN()
        clust = labeled_vectors.cluster_labeled_vectors(method=scan)
        with pytest.raises(ValueError):
            clust.plot_clustered()

        with pytest.raises(ValueError):
            labeled_vectors.plot_clustered()

    def test_plot_cluster_labeled_vectors(self):
        vectors = np.reshape(np.arange(1000), (100, 10))
        labels = np.repeat(np.arange(10), 10)
        vectors = np.hstack((vectors, labels[:, np.newaxis]))
        columns = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "label"]
        vectors[:50, 0] = 10
        vectors[:50, 1] = 10
        vectors[50:, 0] = 20
        vectors[50:, 1] = 20
        labeled_vectors = LabeledDiffractionVectors2D(
            data=vectors, column_names=columns
        )
        scan = DBSCAN()
        clust = labeled_vectors.cluster_labeled_vectors(method=scan)
        clust.plot_clustered()

        clust.plot_clustered(labels=[0, 1])
        s = ElectronDiffraction2D(np.zeros((10, 10)))
        clust.plot_clustered(signal=s)

    def test_only_signal_axes_error(self):
        vectors = np.reshape(np.arange(1000), (100, 10))
        labels = np.repeat(np.arange(10), 10)
        vectors = np.hstack((vectors, labels[:, np.newaxis]))
        vects = np.stack([vectors, vectors])
        columns = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "label"]
        labeled_vectors = LabeledDiffractionVectors2D(data=vects, column_names=columns)
        with pytest.raises(ValueError):
            labeled_vectors.cluster_labeled_vectors(method=DBSCAN())



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Utils for Correlations."""

import numpy as np
from pyxem.utils._deprecated import deprecated


def _correlation(z, axis=0, mask=None, wrap=True, normalize=True):
    r"""A generic function for applying a correlation with a mask.

    Takes a nd image and then preforms a auto-correlation on some axis.
    Used in the electron correlation and angular correlation codes. Uses
    the fft to speed up the correlation.

    Parameters
    ----------
    z: np.array
        A nd numpy array
    axis: int
        The axis to apply the correlation to
    mask: np.array
        A boolean array of the same size as z
    wrap: bool
        Allow the function to wrap or add zeros to the beginning and the
        end of z on the specified axis
    normalize: bool
        Subtract <I(\theta)>^2 and divide by <I(\theta)>^2
    """
    if not wrap:
        z_shape = np.shape(z)
        padder = [(0, 0)] * len(z_shape)
        pad = z_shape[axis]  # This will be faster if the length of the axis
        # is a power of 2.  Based on the numpy implementation.  Not terribly
        # faster I think..
        padder[axis] = (pad, pad)
        slicer = [
            slice(None),
        ] * len(z_shape)
        slicer[axis] = slice(0, -2 * pad)  # creating the proper slices
        if mask is None:
            mask = np.zeros(shape=np.shape(z))
        z = np.pad(z, padder, "constant")

    if mask is not None or not wrap:  # Need to scale for wrapping properly
        m = np.array(mask, dtype=bool)  # make sure it is a boolean array
        # this is to determine how many of the variables were non zero... This is really dumb.  but...
        # it works and I should stop trying to fix it (wreak it)
        mask_boolean = ~m  # inverting the boolean mask
        if not wrap:  # padding with zeros to the function along some axis
            m = np.pad(
                m, padder, "constant"
            )  # all the zeros are masked (should account for padding
            #  when normalized.
            mask_boolean = np.pad(mask_boolean, padder, "constant")
        mask_fft = np.fft.rfft(mask_boolean, axis=axis)
        number_unmasked = np.fft.irfft(
            mask_fft * np.conjugate(mask_fft), axis=axis
        ).real
        number_unmasked[number_unmasked < 1] = (
            1  # get rid of divide by zero error for completely masked rows
        )
        z[m] = 0

    # fast method uses a FFT and is a process which is O(n) = n log(n)
    I_fft = np.fft.rfft(z, axis=axis)
    a = np.fft.irfft(I_fft * np.conjugate(I_fft), axis=axis)

    if mask is not None:
        a = np.multiply(np.divide(a, number_unmasked), np.shape(z)[0])

    if normalize:  # simplified way to calculate the normalization
        row_mean = np.mean(a, axis=axis)
        row_mean[row_mean == 0] = 1
        row_mean = np.expand_dims(row_mean, axis=axis)
        a = np.divide(np.subtract(a, row_mean), row_mean)

    if not wrap:
        a = a[tuple(slicer)]
    return a


def _power(z, axis=0, mask=None, wrap=True, normalize=True):
    """The power spectrum of the correlation.

    This method is a little more complex if mask is not None due to
    the extra calculations necessary to ignore some of the pixels
    during the calculations.

    Parameters
    ----------
    z: np.array
        Some n-d array to get the power spectrum from.
    axis: int
        The axis to preform the operation on.
    mask: np.array
        A boolean mask to be applied.
    wrap: bool
        Choose if the function should wrap.  In most cases this will be True
        when calculating the power of some function
    normalize: bool
        Choose to normalize the function by the mean.

    Returns
    -------
    power: np.array
        The power spectrum along some axis
    """
    if mask is None:  # This might not normalize things as well
        I_fft = np.fft.rfft(z, axis=axis)
        return (I_fft * np.conjugate(I_fft)).real
    else:
        return np.power(
            np.fft.rfft(
                _correlation(z=z, axis=axis, mask=mask, wrap=wrap, normalize=normalize)
            ),
            2,
        ).real


def _pearson_correlation(z, mask=None, mode="full"):
    """
    Calculate Pearson cross-correlation of the image with itself
    after rotation as a function of rotation

     Parameters
    ----------
    z: np.array
        Input image in 2D array
    mask: np.array
        A boolean mask to be applied.
    mode: str 'full' or 'kresolved'
        Mode for calculating pearson correlation, default is 'full'

    Returns
    -------
    p_correlation: np.array
        Pearson correlation of the input image

    """
    if mask is not None:
        # this is to determine how many of the elements were unmasked for normalization
        m = np.array(mask, dtype=bool)
        mask_bool = ~m
        mask_fft = np.fft.fft(mask_bool, axis=1)
        n_unmasked = np.fft.ifft(mask_fft * mask_fft.conj()).real
        n_unmasked[n_unmasked < 1] = (
            1  # avoid dividing by zero for completely masked rows
        )
        z[m] = 0  # set masked pixels to zero
        fft_intensity = np.divide(np.fft.fft(z, axis=1), n_unmasked)
        a = np.multiply(
            np.fft.ifft(fft_intensity * fft_intensity.conj()).real, n_unmasked
        )
    else:
        z_length = np.shape(z)[1]
        fft_intensity = np.fft.fft(z, axis=1) / z_length
        a = np.fft.ifft(fft_intensity * fft_intensity.conj(), axis=1).real * z_length

    if mode == "full":
        p_correlation = (np.mean(a, axis=0) - np.mean(z) ** 2) / (
            np.mean(z**2) - np.mean(z) ** 2
        )
    elif mode == "kresolved":
        p_correlation = (a - np.mean(z, axis=1)[:, np.newaxis] ** 2) / (
            np.mean(z**2, axis=1) - np.mean(z, axis=1) ** 2
        )[:, np.newaxis]
    else:
        raise ValueError(
            "Mode cannot be " + mode + ", must be chosen from 'full' or 'kresolved'"
        )

    return p_correlation


def _wrap_set_float(target, bottom, top, value):
    """This function sets values in a list assuming that
    the list is circular and allows for float bottom and float top
    which are equal to the residual times that value.

    Parameters
    ----------
    target: list
        The list or array to be adjusted
    bottom: float
        The bottom index. Can be a float in which case the value will be
        split.  i.e. 7.5 will set target[7]= value * .5
    top: float
        The top index. Can be a float in which case the value will be
        split.  i.e. 7.5 will set target[8]= value * .5
    value:
        The value to set the range as.
    """
    if (top - bottom) > 0:
        value = value / (top - bottom + 1)

    ceiling_bottom = int(np.ceil(bottom))
    residual_bottom = ceiling_bottom - bottom
    floor_top = int(np.floor(top))
    residual_top = top - floor_top
    if floor_top > len(target) - 1:
        target[ceiling_bottom:] = value
        new_floor_top = floor_top % len(target)
        target[new_floor_top] = value
        target[new_floor_top + 1] = value * residual_top
    elif ceiling_bottom < 0:
        target[:floor_top] = value
        target[ceiling_bottom:] = value
        target[ceiling_bottom - 1] = value * residual_bottom
    else:
        target[ceiling_bottom : floor_top + 1] = value
        target[ceiling_bottom - 1] = value * residual_bottom
        if floor_top + 1 > len(target) - 1:
            target[0] = value * residual_top
        else:
            target[floor_top + 1] = value * residual_top
    return target


def _get_interpolation_matrix(angles, angular_range, num_points, method="average"):
    """Returns an interpolation matrix for slicing a dataset based on the given angles as
    well as an angular range.

    The method is separated into two based on if the angles are all treated equally or if
    only the first or max angle is considered for some list of angles.

    Parameters
    ----------
    angles: list
        A list of a list of angles where each row represents a grouping of angles to be considered.
        For example if you are trying to get the expression of 2-fold and 4-fold symmetry this
        would look like [[0,np.pi],[0,np.pi/2,np.pi,np.pi*3/2]]
    angular_range: float
        The angular range in rad to consider.  If zero only the nearest pixel will be considered
    num_points: int
        The number of points in the azimuthal range to consider
    method: str
        One of "average" "first" or "max".  Changes how the interpolation matrix is created
        for further processing

    """
    if method == "average":
        angular_ranges = [
            (angle - angular_range / (2 * np.pi), angle + angular_range / (2 * np.pi))
            for angle in angles
        ]
        angular_ranges = np.multiply(angular_ranges, num_points)
        interpolation_matrix = np.zeros(num_points)
        for i, angle in enumerate(angular_ranges):
            _wrap_set_float(
                interpolation_matrix, top=angle[1], bottom=angle[0], value=1
            )
        return interpolation_matrix
    else:
        angular_ranges = [
            (angle - angular_range / (2 * np.pi), angle + angular_range / (2 * np.pi))
            for angle in angles
        ]
        angular_ranges = np.multiply(angular_ranges, num_points)
        interpolation_matrix = np.zeros((len(angles), num_points))
        for i, angle in enumerate(angular_ranges):
            _wrap_set_float(
                interpolation_matrix[i, :], top=angle[1], bottom=angle[0], value=1
            )
        return interpolation_matrix


def _symmetry_stem(signal, interpolation, method="average"):
    """Returns the "average" "max" or "first" value for some given signal and an interpolation matrix.

    The interpolation matrix is defined by the  `_get_interpolation_matrix` function which creates a
    matrix which when matrix multiplied by the signal returns the "average", "max" or "first" value for
    certain angles related to some symmetry operation.

    ie. For 4-Fold symmetry operations the angles are [0, pi/2, pi,3pi/2]

    Parameters
    ----------
    signal: np.array
        The signal from which to calculate the symmetry stem given an interpolation matrix
    interpolation: np.array
        The interpolation matrix for calculating the expression of certain symmetry operations.
    method:str
        One of "average", "max" or "first"
    """
    if method == "average":
        return np.matmul(signal, np.transpose(interpolation))
    elif method == "max":
        val = np.transpose(
            [
                np.amax([np.matmul(signal, np.transpose(i)) for i in interp], axis=0)
                for interp in interpolation
            ]
        )
    elif method == "first":
        val = np.transpose(
            [np.matmul(signal, np.transpose(interp[0])) for interp in interpolation]
        )
    else:
        raise ValueError("Method must be one of `average`, `max` or `first`")
    return val


@deprecated(since="0.18.0", removal="0.20.0")
def corr_to_power(z):
    return np.power(np.fft.rfft(z, axis=1), 2).real


def _corr_to_power(z):
    return np.power(np.fft.rfft(z, axis=1), 2).real



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Utils for operating on pixelated signals."""

import copy
import numpy as np
import scipy.ndimage as ndi
from hyperspy.misc.utils import isiterable


def _threshold_and_mask_single_frame(im, threshold=None, mask=None):
    image = copy.deepcopy(im)
    if mask is not None:
        image *= mask
    if threshold is not None:
        mean_value = ndi.mean(image, mask) * threshold
        image[image <= mean_value] = 0
        image[image > mean_value] = 1
    return image


def _shift_single_frame(im, shift_x, shift_y, interpolation_order=1):
    im_shifted = ndi.shift(im, (-shift_y, -shift_x), order=interpolation_order)
    return im_shifted


def _make_circular_mask(centerX, centerY, imageSizeX, imageSizeY, radius):
    """Make a circular mask in a bool array for masking a region in an image.

    Parameters
    ----------
    centreX, centreY : float
        Centre point of the mask.
    imageSizeX, imageSizeY : int
        Size of the image to be masked.
    radius : float
        Radius of the mask.

    Returns
    -------
    Boolean Numpy 2D Array
        Array with the shape (imageSizeX, imageSizeY) with the mask.

    Examples
    --------
    >>> import numpy as np
    >>> import pyxem.utils._pixelated_stem_tools as pst
    >>> image = np.ones((9, 9))
    >>> mask = pst._make_circular_mask(4, 4, 9, 9, 2)
    >>> image_masked = image*mask
    >>> import matplotlib.pyplot as plt
    >>> cax = plt.imshow(image_masked)
    """
    x, y = np.ogrid[-centerY : imageSizeY - centerY, -centerX : imageSizeX - centerX]
    mask = x * x + y * y <= radius * radius
    return mask


def _find_longest_distance(
    imX,
    imY,
    centreX_min,
    centreY_min,
    centreX_max,
    centreY_max,
):
    max_value = max(
        int(((imX - centreX_min) ** 2 + (imY - centreY_min) ** 2) ** 0.5),
        int(((centreX_max) ** 2 + (imY - centreY_min) ** 2) ** 0.5),
        int(((imX - centreX_min) ** 2 + (centreY_max) ** 2) ** 0.5),
        int((centreX_max**2 + centreY_max**2) ** 0.5),
    )
    return max_value


def _make_centre_array_from_signal(signal, x=None, y=None):
    a_m = signal.axes_manager
    shape = a_m.navigation_shape[::-1]
    if x is None:
        centre_x_array = np.ones(shape) * a_m.signal_axes[0].value2index(0)
    else:
        centre_x_array = np.ones(shape) * x
    if y is None:
        centre_y_array = np.ones(shape) * a_m.signal_axes[1].value2index(0)
    else:
        centre_y_array = np.ones(shape) * y
    if not isiterable(centre_x_array):
        centre_x_array = np.array([centre_x_array])
    if not isiterable(centre_y_array):
        centre_y_array = np.array([centre_y_array])
    return (centre_x_array, centre_y_array)


def _get_radial_profile_of_diff_image(
    diff_image, centre_x, centre_y, normalize, radial_array_size, mask=None
):
    """Radially average a single diffraction image around a centre position.

    Radially profiles the data, averaging the intensity in rings
    out from the centre. Unreliable as we approach the edges of the
    image as it just profiles the corners. Less pixels there so become
    effectively zero after a certain point.

    Parameters
    ----------
    diff_image : 2-D numpy array
        Array consisting of a single diffraction image.
    centre_x : number
        Centre x position of the diffraction image.
    centre_y : number
        Centre y position of the diffraction image.
    radial_array_size : number
    mask : numpy bool array, optional
        Mask parts of the diffraction image, regions where
        the mask is True will be included in the radial profile.

    Returns
    -------
    1-D numpy array of the radial profile.

    """
    radial_array = np.zeros(shape=radial_array_size, dtype=np.float64)
    y, x = np.indices((diff_image.shape))
    r = np.sqrt((x - centre_x) ** 2 + (y - centre_y) ** 2)
    r = r.astype(int)
    if mask is None:
        r_flat = r.ravel()
        diff_image_flat = diff_image.ravel()
    else:
        r_flat = r[mask].ravel()
        diff_image_flat = diff_image[mask].ravel()
    tbin = np.bincount(r_flat, diff_image_flat)
    nr = np.bincount(r_flat)
    nr.clip(1, out=nr)  # To avoid NaN in data due to dividing by 0
    if normalize:
        radial_profile = tbin / nr
    else:
        radial_profile = tbin
    radial_array[0 : len(radial_profile)] = radial_profile
    return radial_array


def _get_angle_sector_mask(
    signal, angle0, angle1, centre_x_array=None, centre_y_array=None
):
    """Get a bool array with True values between angle0 and angle1.
    Will use the (0, 0) point as given by the signal as the centre,
    giving an "angular" slice. Useful for analysing anisotropy in
    diffraction patterns.

    Parameters
    ----------
    signal : HyperSpy 2-D signal
        Can have several navigation dimensions.
    angle0, angle1 : numbers

    Returns
    -------
    Mask : Numpy array
        The True values will be the region between angle0 and angle1.
        The array will have the same dimensions as the input signal.

    Examples
    --------
    >>> import numpy as np
    >>> import pyxem.utils._pixelated_stem_tools as pst
    >>> s = pxm.signals.Diffraction2D(np.arange(100).reshape(10, 10))
    >>> s.axes_manager.signal_axes[0].offset = -5
    >>> s.axes_manager.signal_axes[1].offset = -5
    >>> mask = pst._get_angle_sector_mask(s, 0.5*np.pi, np.pi)

    """
    if angle0 > angle1:
        raise ValueError(
            "angle1 ({0}) needs to be larger than angle0 ({1})".format(angle1, angle0)
        )

    bool_array = np.zeros_like(signal.data, dtype=bool)
    for s in signal:
        indices = signal.axes_manager.indices[::-1]
        signal_axes = s.axes_manager.signal_axes
        if centre_x_array is not None:
            if indices == ():
                signal_axes[0].offset = -centre_x_array[0]
            else:
                signal_axes[0].offset = -centre_x_array[indices]
        if centre_y_array is not None:
            if indices == ():
                signal_axes[1].offset = -centre_y_array[0]
            else:
                signal_axes[1].offset = -centre_y_array[indices]
        x_size = signal_axes[1].size * 1j
        y_size = signal_axes[0].size * 1j
        x, y = np.mgrid[
            signal_axes[1].low_value : signal_axes[1].high_value : x_size,
            signal_axes[0].low_value : signal_axes[0].high_value : y_size,
        ]
        t = np.arctan2(x, y) + np.pi
        if (angle1 - angle0) >= (2 * np.pi):
            bool_array[indices] = True
        else:
            angle0 = angle0 % (2 * np.pi)
            angle1 = angle1 % (2 * np.pi)
            if angle0 < angle1:
                bool_array[indices] = (t > angle0) * (t < angle1)
            elif angle1 < angle0:
                bool_array[indices] = (t > angle0) * (t <= (2 * np.pi))
                bool_array[indices] += (t >= 0) * (t < angle1)
    return bool_array


def _copy_signal2d_axes_manager_metadata(signal_original, signal_new):
    ax_o = signal_original.axes_manager.signal_axes
    ax_n = signal_new.axes_manager.signal_axes
    _copy_axes_object_metadata(ax_o[0], ax_n[0])
    _copy_axes_object_metadata(ax_o[1], ax_n[1])


def _copy_axes_object_metadata(axes_original, axes_new):
    axes_new.scale = axes_original.scale
    axes_new.offset = axes_original.offset
    axes_new.name = axes_original.name
    axes_new.units = axes_original.units


def _copy_signal_all_axes_metadata(signal_original, signal_new):
    if signal_original.axes_manager.shape != signal_new.axes_manager.shape:
        raise ValueError(
            "signal_original and signal_new must have the same shape, not "
            "{0} and {1}".format(
                signal_original.axes_manager.shape, signal_new.axes_manager.shape
            )
        )
    for iax in range(len(signal_original.axes_manager.shape)):
        ax_o = signal_original.axes_manager[iax]
        ax_n = signal_new.axes_manager[iax]
        _copy_axes_object_metadata(ax_o, ax_n)



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Utils for polar 2D Diffraction Pattern transformations."""

import numpy as np
from pyxem.utils.diffraction import find_beam_center_blur
from scipy import ndimage
from pyxem.utils.cuda_utils import get_array_module


try:
    import cupy as cp
    import cupyx.scipy.ndimage as ndigpu

    CUPY_INSTALLED = True
except ImportError:
    CUPY_INSTALLED = False
    cp = None
    ndigpu = None


""" These are designed to be fast and used for indexation, for data correction, see radial_utils"""


def _cartesian_positions_to_polar_nonround(x, y, delta_r=1.0, delta_theta=1.0):
    r = np.sqrt(x**2 + y**2) / delta_r
    theta = np.mod(np.rad2deg(np.arctan2(y, x)), 360) / delta_theta
    return r, theta


def _cartesian_positions_to_polar(x, y, delta_r=1, delta_theta=1):
    """
    Convert 2D cartesian image coordinates to polar image coordinates

    Parameters
    ----------
    x : 1D numpy.ndarray
        x coordinates
    y : 1D numpy.ndarray
        y coordinates
    delta_r : float
        sampling interval in the r direction
    delta_theta : float
        sampling interval in the theta direction

    Returns
    -------
    r : 1D numpy.ndarray
        r coordinate or x coordinate in the polar image
    theta : 1D numpy.ndarray
        theta coordinate or y coordinate in the polar image
    """
    r, theta = _cartesian_positions_to_polar_nonround(x, y, delta_r, delta_theta)
    r = np.rint(r).astype(np.int32)
    theta = np.rint(theta).astype(np.int32)
    return r, theta


def get_template_polar_coordinates(
    simulation,
    in_plane_angle=0.0,
    delta_r=1,
    delta_theta=1,
    max_r=None,
    mirrored=False,
    rounded=True,
):
    """
    Convert a single simulation to polar coordinates

    Parameters
    ----------
    simulation : diffsims.sims.diffraction_simulation.DiffractionSimulation
        The diffraction pattern template object
    in_plane_angle : float, optional
        In-plane angle in degrees that the template should be rotated by
    delta_r : float, optional
        Scaling factor for r in case the points should align with a rescaled
        polar image. If delta_r = 1 (default), then r is in units of pixels
        of the original image.
    delta_theta : float, optional
        Scaling factor for theta in case the points should align with a
        rescaled polar image. If delta_theta = 1 (default), then theta is in
        units of degrees.
    max_r : float, optional
        Maximum radial distance to consider. In units of pixels, not scaled
        by delta_r = the width of the polar image.
    mirrored : bool, optional
        Whether to mirror the template
    rounded: bool, optional
        Whether the coordinates should be rounded to integers

    Returns
    -------
    r : np.ndarray
        The r coordinates of the diffraction spots in the template scaled by
        delta_r
    theta : np.ndarray
        The theta coordinates of the diffraction spots in the template, scaled
        by delta_theta
    intensities : np.ndarray
        The intensities of the diffraction spots with dtype float
    """
    x = simulation.calibrated_coordinates[:, 0]
    y = simulation.calibrated_coordinates[:, 1]
    intensities = simulation.intensities.astype(np.float64)
    r, theta = _cartesian_positions_to_polar_nonround(x, y, delta_r, delta_theta)
    if max_r is not None:
        condition = r < max_r
        r = r[condition]
        theta = theta[condition]
        intensities = intensities[condition]
    theta = np.mod(theta + in_plane_angle / delta_theta, 360 / delta_theta)
    if mirrored:
        theta = 360 / delta_theta - theta
    if rounded:
        theta = theta.astype(np.int32)
        r = r.astype(np.int32)
    return r, theta, intensities


def get_template_cartesian_coordinates(
    simulation,
    center=(0.0, 0.0),
    in_plane_angle=0.0,
    window_size=None,
    mirrored=False,
):
    """
    Get the cartesian coordinates of the diffraction spots in a template

    Parameters
    ----------
    simulation : diffsims.sims.diffraction_simulation.DiffractionSimulation
        The diffraction pattern
    center : 2-tuple of ints, optional
        The coordinate of the direct beam in pixel coordinates
    in_plane_angle : float, optional
        Angle in degrees representing an in-plane rotation of the template
        around the direct beam
    window_size : 2-tuple, optional
        Only return the reflections within the (width, height) in pixels
        of the image
    mirrored : bool, optional
        Whether to mirror the template

    Returns
    -------
    x : np.ndarray
        x coordinates of the diffraction spots in the template in pixel units
    y : np.ndarray
        y coordinates of the diffraction spots in the template in pixel units
    intensities: np.ndarray
        intensities of the spots
    """
    ox = simulation.calibrated_coordinates[:, 0]
    oy = simulation.calibrated_coordinates[:, 1]
    if mirrored:
        oy = -oy
    intensities = simulation.intensities
    c = np.cos(np.deg2rad(in_plane_angle))
    s = np.sin(np.deg2rad(in_plane_angle))
    # rotate it
    x = c * ox - s * oy + center[0]
    y = s * ox + c * oy + center[1]
    if window_size is not None:
        condition = (x < window_size[0]) & (y < window_size[1]) & (y >= 0) & (x >= 0)
        x = x[condition]
        y = y[condition]
        intensities = intensities[condition]
    return x, y, intensities


def get_polar_pattern_shape(image_shape, delta_r, delta_theta, max_r=None):
    """
    Returns the shape of images if they would be transformed
    to polar coordinates.

    Parameters
    ----------
    image_shape: 2-Tuple
        (height, width) of the images in cartesian coordinates
    delta_r : float
        size of pixels in the r-direction, in units of cartesian pixels.
    delta_theta : float
        size of pixels in the theta-direction, in degrees
    max_r : float, optional
        maximum radius of the polar image, by default the cartesian distance
        in pixels from the center of the image to the corner, rounded up to
        the nearest integer.

    Returns
    -------
    theta_dim : int
        size of the theta dimension
    r_dim : int
        size of the r dimension
    """
    if max_r is None:
        half_y = image_shape[0] / 2
        half_x = image_shape[1] / 2
        r_dim = int(np.ceil(np.sqrt(half_x**2 + half_y**2)) / delta_r)
    else:
        r_dim = int(round(max_r / delta_r))
    theta_dim = int(round(360 / delta_theta))
    return (theta_dim, r_dim)


def _get_map_function(dispatcher):
    return ndimage.map_coordinates if dispatcher == np else ndigpu.map_coordinates


def _warp_polar_custom(
    image, center, radius, output_shape, order=1, precision=np.float64
):
    """
    Function to emulate warp_polar in skimage.transform on both CPU and GPU. Not all
    parameters are supported

    Parameters
    ----------
    image: numpy.ndarray or cupy.ndarray
        Input image. Only 2-D arrays are accepted.
    center: tuple (row, col)
        Point in image that represents the center of the transformation
        (i.e., the origin in cartesian space). Values can be of type float.
    radius: float
        Radius of the circle that bounds the area to be transformed.
    output_shape: tuple (row, col)
        Shape of the output polar image
    order: int, optional
        Order of interpolation between pixels
    precision: np.float64 or np.float32
        Double or single precision output

    Returns
    -------
    polar: numpy.ndarray or cupy.ndarray
        polar image of dtype float64

    Notes
    -----
    Speed gains on the GPU will depend on the size of your problem.
    On a Tesla V100 a 10x speed up was achieved with a 256x256 image:
    from 5 ms to 400 microseconds. For a 4000x4000 image a 1000x speed up
    was achieved: from 180 ms to 400 microseconds. However, this does not
    count the time to transfer data from the CPU to the GPU and back.
    """
    dispatcher = get_array_module(image)
    cy, cx = center
    H = output_shape[0]
    W = output_shape[1]
    T = dispatcher.linspace(0, 2 * dispatcher.pi, H).reshape(H, 1)
    R = dispatcher.linspace(0, radius, W).reshape(1, W)
    X = R * dispatcher.cos(T) + cx
    Y = -R * dispatcher.sin(T) + cy
    coordinates = dispatcher.stack([Y, X])
    map_function = _get_map_function(dispatcher)
    polar = map_function(image.astype(precision), coordinates, order=order)
    return polar


def image_to_polar(
    image,
    delta_r=1.0,
    delta_theta=1.0,
    max_r=None,
    find_direct_beam=False,
    direct_beam_position=None,
    order=1,
):
    """Convert a single image to polar coordinates including the option to
    find the direct beam and take this as the center.

    Parameters
    ----------
    image : 2D numpy.ndarray or 2D cupy.ndarray
        Experimental image
    delta_r : float, optional
        The radial increment, determines how many columns will be in the polar
        image.
    delta_theta : float, optional
        The angular increment, determines how many rows will be in the polar
        image
    max_r : float, optional
        The maximum radial distance to include, in units of pixels. By default
        this is the distance from the center of the image to a corner in pixel
        units and rounded up to the nearest integer
    find_direct_beam : bool
        Whether to roughly find the direct beam using `find_beam_center_blur`
        using a gaussian smoothing with sigma = 1. If False then the middle of
        the image will be the center for the polar transform.
    direct_beam_position : 2-tuple
        (x, y) position of the central beam in pixel units. This overrides
        the automatic find_maximum parameter if set to True.
    order : int
        Order of interpolation for the polar transform

    Returns
    -------
    polar_image : 2D numpy.ndarray or cupy.ndarray
        Array representing the polar transform of the image with shape
        (360/delta_theta, max_r/delta_r)
    """
    half_x, half_y = image.shape[1] / 2, image.shape[0] / 2
    if direct_beam_position is not None:
        c_x, c_y = direct_beam_position
    elif find_direct_beam:
        c_x, c_y = find_beam_center_blur(image, 1)
    else:
        c_x, c_y = half_x, half_y
    output_shape = get_polar_pattern_shape(
        image.shape, delta_r, delta_theta, max_r=max_r
    )
    maximum_radius = output_shape[1] * delta_r
    return _warp_polar_custom(
        image,
        center=(c_y, c_x),
        output_shape=output_shape,
        radius=maximum_radius,
        order=order,
    )


def _chunk_to_polar(
    images,
    center,
    radius,
    output_shape,
    precision=np.float64,
):
    """
    Convert a chunk of images to polar coordinates

    Parameters
    ----------
    images : (scan_x, scan_y, x, y) np.ndarray or cp.ndarray
        diffraction patterns
    center : 2-Tuple of float
        center of the images in pixels (row, col) = (c_y, c_x)
    radius : float
        maximum radius to consider in the image. This gets mapped onto
        output_shape[1]
    output_shape : 2-Tuple of int
        (height, width) of the output polar images
    precision : np.float32 or np.float64

    Returns
    -------
    polar_chunk : (scan_x, scan_y, max_r/delta_theta, 360/delta_theta) np.ndarray or cp.ndarray
        diffraction patterns in polar coordinates. Returns a cuda or numpy array depending
        on the device
    """
    dispatcher = get_array_module(images)
    polar_chunk = dispatcher.empty(
        (images.shape[0], images.shape[1], output_shape[0], output_shape[1]),
        dtype=precision,
    )
    for idx, idy in np.ndindex(images.shape[:-2]):
        polar_chunk[idx, idy] = _warp_polar_custom(
            images[idx, idy],
            center=center,
            radius=radius,
            output_shape=output_shape,
        )
    return polar_chunk



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""Utils for using dask."""

import numpy as np
import dask.array as da
import scipy.ndimage as ndi
from skimage import morphology
from hyperspy.misc.utils import isiterable


def _align_single_frame(image, shifts, **kwargs):
    temp_image = ndi.shift(image, shifts[::-1], **kwargs)
    return temp_image


def _get_signal_dimension_chunk_slice_list(chunks):
    """Convenience function for getting the signal chunks as slices

    The slices are assumed to be used on a HyperSpy signal object.
    Thus the input will be in the Dask chunk order (y, x), while the
    output will be in the HyperSpy order (x, y).

    """
    chunk_slice_raw_list = da.core.slices_from_chunks(chunks[-2:])
    chunk_slice_list = []
    for chunk_slice_raw in chunk_slice_raw_list:
        chunk_slice_list.append((chunk_slice_raw[1], chunk_slice_raw[0]))
    return chunk_slice_list


def _get_signal_dimension_host_chunk_slice(x, y, chunks):
    chunk_slice_list = _get_signal_dimension_chunk_slice_list(chunks)
    for chunk_slice in chunk_slice_list:
        x_slice, y_slice = chunk_slice
        if y_slice.start <= y < y_slice.stop:
            if x_slice.start <= x < x_slice.stop:
                return chunk_slice
    return False


def _intensity_peaks_image_single_frame(frame, peaks, disk_r):
    """Intensity of the peaks is calculated by taking the mean value
    of the pixel values inside radius disk_r where the centers are the
    peak positions. If the peak position plus disk_r exceed the detector
    edges, then the intensity for that peak will be put to zero.

    Parameters
    ----------
    frame : NumPy 2D array
    peaks: Numpy 2D array with x and y coordinates of peaks
    disk : NumPy 2D array
        Must be smaller than frame
        peaks: NumPy Object
        can have multiple peaks per image

    Returns
    -------
    intensity_array : NumPy array with
        peak coordinates and intensity of peaks

    Examples
    --------
    >>> import pyxem.utils.dask_tools as dt
    >>> s = pxm.data.dummy_data.dummy_data.get_cbed_signal()
    >>> peaks = np.array(([50,50],[25,50]))
    >>> intensity = dt._intensity_peaks_image_single_frame(
    ...     s.data[0,0,:,:], peaks, 5)

    """
    array_shape = peaks.shape
    mask = morphology.disk(disk_r)
    size = np.shape(frame)
    intensity_array = np.zeros((array_shape[0], 3), dtype="float64")
    for i in range(array_shape[0]):
        cx = int(peaks[i, 0])
        cy = int(peaks[i, 1])
        intensity_array[i, 0] = peaks[i, 0]
        intensity_array[i, 1] = peaks[i, 1]
        if (
            (cx - disk_r < 0)
            | (cx + disk_r + 1 >= size[0])
            | (cy - disk_r < 0)
            | (cy + disk_r + 1 >= size[1])
        ):
            intensity_array[i, 2] = 0
        else:
            subframe = frame[
                cx - disk_r : cx + disk_r + 1, cy - disk_r : cy + disk_r + 1
            ]
            intensity_array[i, 2] = np.mean(mask * subframe)

    return intensity_array


def _get_chunking(signal, chunk_shape=None, chunk_bytes=None):
    """Get chunk tuple based on the size of the dataset.

    The signal dimensions will be within one chunk, and the navigation
    dimensions will be chunked based on either chunk_shape, or
    be optimized based on the chunk_bytes.

    Parameters
    ----------
    signal : hyperspy or pyxem signal
    chunk_shape : int, optional
        Size of the navigation chunk, of None (the default), the chunk
        size will be set automatically.
    chunk_bytes : int or string, optional
        Number of bytes in each chunk. For example '60MiB'. If None (the default),
        the limit will be '30MiB'. Will not be used if chunk_shape is None.

    Returns
    -------
    chunks : tuple
    """
    if chunk_bytes is None:
        chunk_bytes = "30MiB"
    nav_dim = signal.axes_manager.navigation_dimension
    sig_dim = signal.axes_manager.signal_dimension
    if chunk_shape is not None:
        if not isiterable(chunk_shape):
            chunk_shape = [chunk_shape] * nav_dim

    chunks_dict = {}
    for i in range(nav_dim):
        if chunk_shape is None:
            chunks_dict[i] = "auto"
        else:
            chunks_dict[i] = chunk_shape[i]
    for i in range(nav_dim, nav_dim + sig_dim):
        chunks_dict[i] = -1

    chunks = da.core.normalize_chunks(
        chunks=chunks_dict,
        shape=signal.data.shape,
        limit=chunk_bytes,
        dtype=signal.data.dtype,
    )
    return chunks


def _get_dask_array(signal, chunk_shape=None, chunk_bytes=None):
    if signal._lazy:
        dask_array = signal.data
    else:
        chunks = _get_chunking(signal, chunk_shape, chunk_bytes)
        dask_array = da.from_array(signal.data, chunks=chunks)
    return dask_array



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
import hyperspy.api as hs
import dask.array as da

from hyperspy._signals.signal2d import Signal2D
from diffsims.libraries.vector_library import DiffractionVectorLibrary
from diffsims.sims.diffraction_simulation import ProfileSimulation
from diffsims.generators.diffraction_generator import DiffractionGenerator
from diffsims.generators.library_generator import DiffractionLibraryGenerator
from diffsims.libraries.structure_library import StructureLibrary

from pyxem.generators import (
    IndexationGenerator,
    TemplateIndexationGenerator,
    ProfileIndexationGenerator,
    VectorIndexationGenerator,
    AcceleratedIndexationGenerator,
)
from pyxem.signals import (
    ElectronDiffraction2D,
    DiffractionVectors,
)
from pyxem.utils.indexation_utils import OrientationResult
from unittest.mock import Mock
import sys


def generate_library(good_library):
    """Here we're testing the __init__ so we focus on 0 being the first entry of the orientations"""
    mock_sim_1 = Mock()
    mock_sim_1.calibrated_coordinates = np.array(
        [
            [0, 1, 0],
            [1, 0, 0],
        ]
    )
    mock_sim_1.intensities = np.array(
        [
            2,
            3,
        ]
    )
    simlist = [mock_sim_1, mock_sim_1]

    lead_number = 0 if good_library else 1
    orientations = np.array(
        [
            [0, 2, 3],
            [lead_number, 4, 5],
        ]
    )
    library = {}
    library["dummyphase"] = {"simulations": simlist, "orientations": orientations}
    return library


def test_old_indexer_routine():
    with pytest.raises(ValueError):
        _ = IndexationGenerator("a", "b")
    with pytest.raises(ValueError):
        _ = TemplateIndexationGenerator("a", "b")


@pytest.mark.skipif(sys.platform == "darwin", reason="Fails on Mac OSX")
@pytest.mark.slow
@pytest.mark.parametrize("good_library", [True, False])
def test_AcceleratedIndexationGenerator(good_library):
    signal = ElectronDiffraction2D((np.ones((2, 2, 256, 256)))).as_lazy()
    library = generate_library(good_library=good_library)

    if good_library:
        acgen = AcceleratedIndexationGenerator(signal, library)
        d = acgen.correlate(n_largest=2)

    elif not good_library:
        with pytest.raises(ValueError):
            acgen = AcceleratedIndexationGenerator(signal, library)

    return None


@pytest.fixture
def profile_simulation():
    return ProfileSimulation(
        magnitudes=[
            0.31891931643691351,
            0.52079306292509475,
            0.6106839974876449,
            0.73651261277849378,
            0.80259601243613932,
            0.9020400452156796,
            0.95675794931074043,
            1.0415861258501895,
            1.0893168446141808,
            1.1645286909108374,
            1.2074090451670043,
            1.2756772657476541,
        ],
        intensities=np.array(
            [
                100.0,
                99.34619104,
                64.1846346,
                18.57137199,
                28.84307971,
                41.31084268,
                23.42104951,
                13.996264,
                24.87559364,
                20.85636003,
                9.46737774,
                5.43222307,
            ]
        ),
        hkls=[
            {(1, 1, 1): 8},
            {(2, 2, 0): 12},
            {(3, 1, 1): 24},
            {(4, 0, 0): 6},
            {(3, 3, 1): 24},
            {(4, 2, 2): 24},
            {(3, 3, 3): 8, (5, 1, 1): 24},
            {(4, 4, 0): 12},
            {(5, 3, 1): 48},
            {(6, 2, 0): 24},
            {(5, 3, 3): 24},
            {(4, 4, 4): 8},
        ],
    )


def test_profile_indexation_generator_init(profile_simulation):
    pig = ProfileIndexationGenerator(
        magnitudes=[
            0.31891931643691351,
            0.52079306292509475,
            0.6106839974876449,
            0.73651261277849378,
            0.80259601243613932,
            0.9020400452156796,
            0.95675794931074043,
            1.0415861258501895,
            1.0893168446141808,
            1.1645286909108374,
            1.2074090451670043,
            1.2756772657476541,
        ],
        simulation=profile_simulation,
    )
    assert isinstance(pig, ProfileIndexationGenerator)


def test_profile_indexation_generator_single_indexation(profile_simulation):
    pig = ProfileIndexationGenerator(
        magnitudes=[
            0.31891931643691351,
            0.52079306292509475,
            0.6106839974876449,
            0.73651261277849378,
            0.80259601243613932,
            0.9020400452156796,
            0.95675794931074043,
            1.0415861258501895,
            1.0893168446141808,
            1.1645286909108374,
            1.2074090451670043,
            1.2756772657476541,
        ],
        simulation=profile_simulation,
    )
    indexation = pig.index_peaks(tolerance=0.02)
    np.testing.assert_almost_equal(indexation[0][0], 0.3189193164369)


def test_vector_indexation_generator_init():
    vectors = DiffractionVectors([[1], [2]])
    vectors.cartesian = [[1], [2]]
    vector_library = DiffractionVectorLibrary()
    vector_indexation_generator = VectorIndexationGenerator(vectors, vector_library)
    assert isinstance(vector_indexation_generator, VectorIndexationGenerator)
    assert vector_indexation_generator.vectors == vectors
    assert vector_indexation_generator.library == vector_library


def test_vector_indexation_generator_cartesian_check():
    vectors = DiffractionVectors([[1], [2]])
    vector_library = DiffractionVectorLibrary()

    with pytest.raises(
        ValueError,
        match="Cartesian coordinates are required in order to index diffraction vectors",
    ):
        vector_indexation_generator = VectorIndexationGenerator(vectors, vector_library)


@pytest.mark.skip(
    reason="Failing test due to using an old version of template matching"
)
def test_vector_indexation_generator_index_vectors(vector_match_peaks, vector_library):
    # vectors not used directly
    vectors = DiffractionVectors(np.array(vector_match_peaks[:, :2]))
    vectors.cartesian = DiffractionVectors(np.array(vector_match_peaks))
    gen = VectorIndexationGenerator(vectors, vector_library)
    indexation = gen.index_vectors(
        mag_tol=0.1, angle_tol=6, index_error_tol=0.3, n_peaks_to_index=2, n_best=5
    )

    # Values are tested directly on the match_vector in the util tests
    assert isinstance(indexation.vectors, DiffractionVectors)

    # (n_best=1, 5 result values from each)
    np.testing.assert_equal(indexation.data.shape, (5,))

    # n_best=1, 3 peaks with hkl)
    np.testing.assert_equal(indexation.hkls.shape, (1, 3, 3))

    refined1 = gen.refine_n_best_orientations(indexation, 1.0, 1.0, n_best=0)

    assert isinstance(refined1.vectors, DiffractionVectors)
    np.testing.assert_equal(refined1.data.shape, (5,))

    refined2 = gen.refine_best_orientation(indexation, 1.0, 1.0)

    assert isinstance(refined2.vectors, DiffractionVectors)
    np.testing.assert_equal(refined2.data.shape, (1,))
    assert isinstance(refined2.data[0], OrientationResult)

    assert refined2.data[0].phase_index == indexation.data[0].phase_index
    assert refined2.data[0].match_rate == indexation.data[0].match_rate

    # Must use a large tolerance here, because there are only 3 vectors
    np.testing.assert_almost_equal(
        np.diag(refined1.data[0].rotation_matrix),
        np.diag(indexation.data[0].rotation_matrix),
        1,
    )
    np.testing.assert_almost_equal(
        np.diag(refined2.data[0].rotation_matrix),
        np.diag(indexation.data[0].rotation_matrix),
        1,
    )



import numpy as np
from orix.crystal_map import Phase
from orix.quaternion import Rotation
from diffpy.structure import Atom, Lattice, Structure
from diffsims.generators.simulation_generator import SimulationGenerator
from pyxem.signals import Diffraction2D
import skimage


def fe_fcc_phase():
    """Create a fe fcc phase with space group 225. This is a fcc structure with
    a lattice parameter of 3.571 Å.

    """
    a = 3.571
    latt = Lattice(a, a, a, 90, 90, 90)
    atom_list = []
    for coords in [[0, 0, 0], [0.5, 0, 0.5], [0, 0.5, 0.5], [0.5, 0.5, 0]]:
        x, y, z = coords[0], coords[1], coords[2]
        atom_list.append(Atom(atype="Fe", xyz=[x, y, z], lattice=latt))  # Motif part A
    struct = Structure(atoms=atom_list, lattice=latt)
    p = Phase(structure=struct, space_group=225)
    return p


def fe_bcc_phase():
    """Create a fe bcc phase with space group 229. This is a bcc structure with
    a lattice parameter of 2.866 Å.

    """
    a = 2.866
    latt = Lattice(a, a, a, 90, 90, 90)
    atom_list = []
    for coords in [[0, 0, 0], [0.5, 0.5, 0.5]]:
        x, y, z = coords[0], coords[1], coords[2]
        atom_list.append(Atom(atype="Fe", xyz=[x, y, z], lattice=latt))  # Motif part A
    struct = Structure(atoms=atom_list, lattice=latt)
    p = Phase(structure=struct, space_group=229)
    return p


def fe_multi_phase_grains(num_grains=2, seed=2, size=20, recip_pixels=128):
    bcc = fe_bcc_phase()
    fcc = fe_fcc_phase()
    gen = SimulationGenerator()
    rotations1 = Rotation.random(num_grains)
    rotations2 = Rotation.random(num_grains)

    sim = gen.calculate_diffraction2d(
        phase=[bcc, fcc],
        rotation=[rotations1, rotations2],
        reciprocal_radius=2.5,
        max_excitation_error=0.1,
    )
    dps_bcc = [
        sim.iphase[0]
        .irot[i]
        .get_diffraction_pattern(
            sigma=3, shape=(recip_pixels, recip_pixels), calibration=0.02
        )
        for i in range(num_grains)
    ]
    dps_fcc = [
        sim.iphase[1]
        .irot[i]
        .get_diffraction_pattern(
            sigma=3, shape=(recip_pixels, recip_pixels), calibration=0.02
        )
        for i in range(num_grains)
    ]
    rng = np.random.default_rng(seed)
    x = rng.integers(0, size, size=num_grains * 2)
    y = rng.integers(0, size, size=num_grains * 2)
    navigator = np.zeros((size, size))
    for i in range(num_grains * 2):
        navigator[x[i], y[i]] = i + 1
    navigator = skimage.segmentation.expand_labels(navigator, distance=size)
    grain_data = np.empty((size, size, recip_pixels, recip_pixels))
    for i in range(num_grains * 2):
        if i < num_grains:
            grain_data[navigator == i + 1] = dps_bcc[i][np.newaxis]
        else:
            grain_data[navigator == i + 1] = dps_fcc[i % num_grains][np.newaxis]
    grains = Diffraction2D(grain_data)
    grains.axes_manager.signal_axes[0].name = "kx"
    grains.axes_manager.signal_axes[1].name = "kx"
    grains.axes_manager.signal_axes[0].units = r"$\AA^{-1}$"
    grains.axes_manager.signal_axes[1].units = r"$\AA^{-1}$"
    grains.axes_manager.signal_axes[0].scale = 0.02
    grains.axes_manager.signal_axes[1].scale = 0.02
    return grains



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.


import pytest
from pytest import approx
import numpy as np
from hyperspy.signals import Signal2D
from pyxem.data.dummy_data import make_diffraction_test_data as mdtd


class TestMakeTestData:
    def test_init(self):
        mdtd.MakeTestData(default=True)
        mdtd.MakeTestData(size_x=1, size_y=10, scale=0.05)
        mdtd.MakeTestData(
            size_x=200,
            size_y=100,
            scale=2,
            default=False,
            blur=False,
            blur_sigma=3.0,
            downscale=False,
        )

    def test_zero_signal(self):
        test_data_1 = mdtd.MakeTestData(default=False)
        assert (test_data_1.signal.data == 0.0).all()

        test_data_2 = mdtd.MakeTestData(default=True)
        test_data_2.set_signal_zero()
        assert (test_data_2.signal.data == 0.0).all()

    def test_repr(self):
        test = mdtd.MakeTestData(size_x=250)
        repr_string = test.__repr__()
        assert str(250) in repr_string


def test_circle_repr():
    xx, yy = np.zeros((2, 10, 10))
    circle = mdtd.Circle(xx, yy, x0=7, y0=5, r=3, intensity=2, scale=1)
    repr_string = circle.__repr__()
    assert str(7) in repr_string


class TestMakeDiffractionTestDataDisks:
    def test_simple_disks(self):
        test0 = mdtd.MakeTestData(
            size_x=100, size_y=100, default=False, blur=False, downscale=False
        )
        test0.add_disk(50, 50, 1, 20)

        assert (test0.signal.isig[49:52, 49:52].data == 20).all()
        test0.signal.data[49:52, 49:52] = 0
        assert not test0.signal.data.any()

        test1 = mdtd.MakeTestData(
            size_x=100, size_y=100, default=False, blur=False, downscale=False
        )
        test1.add_disk(20, 40, 2, 30)
        assert (test1.signal.isig[18:23, 38:43].data == 30).all()
        test1.signal.data[38:43, 18:23] = 0
        assert not test1.signal.data.any()

        test2 = mdtd.MakeTestData(
            size_x=80, size_y=120, default=False, blur=False, downscale=False
        )
        test2.add_disk(50, 40, 1, 10)
        test2.add_disk(20, 30, 2, 30)
        assert (test2.signal.isig[49:52, 39:42].data == 10).all()
        assert (test2.signal.isig[19:22, 29:32].data == 30).all()
        test2.signal.data[39:42, 49:52] = 0
        test2.signal.data[28:33, 18:23] = 0
        assert not test2.signal.data.any()

        test3 = mdtd.MakeTestData(
            size_x=150, size_y=50, default=False, blur=False, downscale=False
        )
        test3.add_disk(200, 400, 2, 30)
        assert not test3.signal.data.any()

        test4 = mdtd.MakeTestData(
            size_x=150, size_y=50, default=False, blur=False, downscale=False
        )
        test4.add_disk(50, 50, 500, 500)
        assert (test4.signal.data == 500).all()

        test5 = mdtd.MakeTestData(
            size_x=100, size_y=200, default=False, blur=False, downscale=True
        )
        test5.add_disk(50, 50, 500, 10)
        assert (test5.signal.data == 10).all()

        test6 = mdtd.MakeTestData(
            size_x=100, size_y=200, default=False, blur=True, downscale=False
        )
        test6.add_disk(50, 30, 500, 10)
        test6_ref = np.full_like(test6.signal.data, 10.0)
        np.testing.assert_allclose(test6.signal.data, test6_ref, rtol=1e-05)

    def test_large_disk(self):
        test_data_1 = mdtd.MakeTestData(size_x=10, size_y=10, scale=0.01, default=False)
        test_data_1.add_disk(x0=5, y0=5, r=20, intensity=100)
        assert (test_data_1.signal.data > 0.0).all()

    def test_repr(self):
        test = mdtd.MakeTestData()
        test.add_disk(x0=70)
        repr_string = test.z_list[0].__repr__()
        assert str(70) in repr_string


class TestMakeDiffractionTestDataRing:
    def test_ring_inner_radius(self):
        r, lw = 20, 2
        x0, y0, scale = 50, 50, 1
        test_data = mdtd.MakeTestData(
            size_x=100, size_y=100, default=False, blur=False, downscale=False
        )
        test_data.add_ring(x0=x0, y0=y0, r=r, intensity=10, lw_pix=lw)
        r_inner = 20 - 2.5 * scale

        s_h0 = test_data.signal.isig[x0, : y0 + 1]
        s_h0_edge = s_h0.axes_manager[0].index2value(s_h0.data[::-1].argmax())
        r_h0 = s_h0_edge - 0.5 * scale
        assert r_h0 == r_inner

        s_h1 = test_data.signal.isig[x0, y0:]
        s_h1_edge = s_h1.axes_manager[0].index2value(s_h1.data.argmax())
        r_h1 = s_h1_edge - 0.5 * scale - 50
        assert r_h1 == r_inner

        s_h2 = test_data.signal.isig[: x0 + 1, y0]
        s_h2_edge = s_h2.axes_manager[0].index2value(s_h2.data[::-1].argmax())
        r_h2 = s_h2_edge - 0.5 * scale
        assert r_h2 == r_inner

        s_h3 = test_data.signal.isig[x0:, y0]
        s_h3_edge = s_h3.axes_manager[0].index2value(s_h3.data.argmax())
        r_h3 = s_h3_edge - 0.5 * scale - 50
        assert r_h3 == r_inner

    def test_ring_outer_radius(self):
        r, lw = 20, 1
        x0, y0, scale = 0, 0, 1
        ring_1 = mdtd.MakeTestData(
            size_x=100, size_y=100, default=False, blur=False, downscale=False
        )
        ring_1.add_ring(x0=x0, y0=y0, r=r, intensity=10, lw_pix=lw)
        s_h0 = ring_1.signal.isig[x0, :]
        s_h0_edge = s_h0.axes_manager[0].index2value(s_h0.data[::-1].argmax())
        r_h0 = s_h0.data.size - s_h0_edge - 0.5 * scale
        r_out = r + lw + 0.5 * scale
        assert r_h0 == r_out
        s_h1 = ring_1.signal.isig[:, y0]
        s_h1_edge = s_h1.axes_manager[0].index2value(s_h1.data[::-1].argmax())
        r_h1 = s_h1.data.size - s_h1_edge - 0.5 * scale
        assert r_h1 == r_out

        r, lw = 20, 1
        x0, y0, scale = 100, 100, 1
        ring_2 = mdtd.MakeTestData(
            size_x=100, size_y=100, default=False, blur=False, downscale=False
        )
        ring_2.add_ring(x0=x0, y0=y0, r=r, intensity=10, lw_pix=lw)
        s_h2 = ring_2.signal.isig[-1, :]
        s_h2_edge = s_h2.axes_manager[0].index2value(s_h2.data.argmax())
        r_h2 = 100 - (s_h2_edge - 0.5 * scale)
        r_out = r + lw + 0.5 * scale
        assert r_h2 == r_out
        s_h3 = ring_2.signal.isig[:, -1]
        s_h3_edge = s_h3.axes_manager[0].index2value(s_h3.data.argmax())
        r_h3 = 100 - (s_h3_edge - 0.5 * scale)
        assert r_h3 == r_out

    def test_ring_radius1(self):
        intensity = 20
        test = mdtd.MakeTestData(
            size_x=100, size_y=100, default=False, blur=False, downscale=False
        )
        test.add_ring(x0=50, y0=50, r=1, intensity=intensity, lw_pix=0)

        assert (test.signal.isig[50, 50].data == 0.0).all()
        test.signal.data[50, 50] = intensity
        assert (test.signal.isig[49:52, 49:52].data == intensity).all()
        test.signal.data[49:52, 49:52] = 0.0
        assert not test.signal.data.any()

    def test_ring_radius2(self):
        intensity = 10
        test = mdtd.MakeTestData(
            size_x=100, size_y=100, default=False, blur=False, downscale=False
        )
        test.add_ring(x0=50, y0=50, r=2, intensity=intensity, lw_pix=0)

        assert (test.signal.isig[49:52, 49:52].data == 0.0).all()
        test.signal.data[49:52, 49:52] = intensity
        assert (test.signal.isig[48:53, 48:53].data == intensity).all()
        test.signal.data[48:53, 48:53] = 0.0
        assert not test.signal.data.any()

    def test_ring_radius_outside_image(self):
        test = mdtd.MakeTestData(
            size_x=100, size_y=100, default=False, blur=False, downscale=False
        )
        test.add_ring(x0=50, y0=50, r=300, intensity=19, lw_pix=1)
        assert not test.signal.data.any()

    def test_ring_rectangle_image(self):
        intensity = 10
        test = mdtd.MakeTestData(
            size_x=100, size_y=50, default=False, blur=False, downscale=False
        )
        test.add_ring(x0=50, y0=25, r=1, intensity=intensity, lw_pix=0)

        assert (test.signal.isig[50, 25].data == 0.0).all()
        test.signal.data[25, 50] = intensity
        assert (test.signal.isig[49:52, 24:27].data == intensity).all()
        test.signal.data[24:27, 49:52] = 0.0
        assert not test.signal.data.any()

    def test_ring_cover_whole_image(self):
        intensity = 10.0
        test = mdtd.MakeTestData(
            size_x=50, size_y=100, default=False, blur=False, downscale=False
        )
        test.add_ring(x0=25, y0=200, r=150, intensity=intensity, lw_pix=100)
        assert (test.signal.data == intensity).all()

    def test_ring_position(self):
        x0, y0, r = 40.0, 60.0, 20
        test = mdtd.MakeTestData(
            size_x=100, size_y=100, default=False, blur=False, downscale=False
        )
        test.add_ring(x0=x0, y0=y0, r=r, intensity=20, lw_pix=0)
        s_h0 = test.signal.isig[x0, 0.0:y0]
        max_h0 = s_h0.axes_manager[0].index2value(s_h0.data.argmax())
        assert max_h0 == y0 - r
        s_h1 = test.signal.isig[x0, y0:]
        max_h1 = s_h1.axes_manager[0].index2value(s_h1.data.argmax())
        assert max_h1 == y0 + r

        s_v0 = test.signal.isig[0:x0, y0]
        max_v0 = s_v0.axes_manager[0].index2value(s_v0.data.argmax())
        assert max_v0 == x0 - r
        s_v1 = test.signal.isig[x0:, y0]
        max_v1 = s_v1.axes_manager[0].index2value(s_v1.data.argmax())
        assert max_v1 == x0 + r

    def test_ring_position_blur(self):
        x0, y0, r = 50, 50, 15
        test = mdtd.MakeTestData(
            size_x=100, size_y=100, default=False, blur=True, downscale=False
        )
        test.add_ring(x0=x0, y0=y0, r=r, intensity=20, lw_pix=1)
        s_h0 = test.signal.isig[x0, 0:y0]
        max_h0 = s_h0.axes_manager[0].index2value(s_h0.data.argmax())
        assert max_h0 == y0 - r
        s_h1 = test.signal.isig[x0, y0:]
        max_h1 = s_h1.axes_manager[0].index2value(s_h1.data.argmax())
        assert max_h1 == y0 + r

        s_v0 = test.signal.isig[0:x0, y0]
        max_v0 = s_v0.axes_manager[0].index2value(s_v0.data.argmax())
        assert max_v0 == x0 - r
        s_v1 = test.signal.isig[x0:, y0]
        max_v1 = s_v1.axes_manager[0].index2value(s_v1.data.argmax())
        assert max_v1 == x0 + r

    def test_ring_position_blur_lw(self):
        x0, y0, r = 50, 50, 15
        test = mdtd.MakeTestData(
            size_x=100, size_y=100, default=False, blur=True, downscale=False
        )
        test.add_ring(x0=x0, y0=y0, r=r, intensity=20, lw_pix=1)
        s_h0 = test.signal.isig[x0, 0:y0]
        max_h0 = s_h0.axes_manager[0].index2value(s_h0.data.argmax())
        assert max_h0 == y0 - r
        s_h1 = test.signal.isig[x0, y0:]
        max_h1 = s_h1.axes_manager[0].index2value(s_h1.data.argmax())
        assert max_h1 == y0 + r

        s_v0 = test.signal.isig[0:x0, y0]
        max_v0 = s_v0.axes_manager[0].index2value(s_v0.data.argmax())
        assert max_v0 == x0 - r
        s_v1 = test.signal.isig[x0:, y0]
        max_v1 = s_v1.axes_manager[0].index2value(s_v1.data.argmax())
        assert max_v1 == x0 + r

    def test_ring_position_downscale(self):
        x0, y0, r = 50, 50, 15
        test = mdtd.MakeTestData(
            size_x=100, size_y=100, default=False, blur=False, downscale=True
        )
        test.add_ring(x0=x0, y0=y0, r=r, intensity=20, lw_pix=0)
        s_h0 = test.signal.isig[x0, 0:y0]
        max_h0 = s_h0.axes_manager[0].index2value(s_h0.data.argmax())
        assert max_h0 == y0 - r
        s_h1 = test.signal.isig[x0, y0:]
        max_h1 = s_h1.axes_manager[0].index2value(s_h1.data.argmax())
        assert max_h1 == y0 + r

        s_v0 = test.signal.isig[0:x0, y0]
        max_v0 = s_v0.axes_manager[0].index2value(s_v0.data.argmax())
        assert max_v0 == x0 - r
        s_v1 = test.signal.isig[x0:, y0]
        max_v1 = s_v1.axes_manager[0].index2value(s_v1.data.argmax())
        assert max_v1 == x0 + r

    def test_ring_position_downscale_and_blur(self):
        x0, y0, r = 50, 50, 18
        test = mdtd.MakeTestData(
            size_x=100, size_y=100, default=False, blur=True, downscale=True
        )
        test.add_ring(x0=x0, y0=y0, r=r, intensity=20, lw_pix=1)
        s_h0 = test.signal.isig[x0, 0:y0]
        max_h0 = s_h0.axes_manager[0].index2value(s_h0.data.argmax())
        assert max_h0 == y0 - r
        s_h1 = test.signal.isig[x0, y0:]
        max_h1 = s_h1.axes_manager[0].index2value(s_h1.data.argmax())
        assert max_h1 == y0 + r

        s_v0 = test.signal.isig[0:x0, y0]
        max_v0 = s_v0.axes_manager[0].index2value(s_v0.data.argmax())
        assert max_v0 == x0 - r
        s_v1 = test.signal.isig[x0:, y0]
        max_v1 = s_v1.axes_manager[0].index2value(s_v1.data.argmax())
        assert max_v1 == x0 + r

    def test_repr(self):
        test = mdtd.MakeTestData()
        test.add_ring(x0=70)
        repr_string = test.z_list[0].__repr__()
        assert str(70) in repr_string

    def test_bad_input_too_wide_ring(self):
        test = mdtd.MakeTestData()
        with pytest.raises(ValueError):
            test.add_ring(r=5, lw_pix=10)


class TestMakeDiffractionTestDataDisksEllipse:
    def test_disk_cover_all(self):
        test = mdtd.MakeTestData(
            size_x=120, size_y=100, default=False, blur=False, downscale=False
        )
        test.add_disk_ellipse(x0=50, y0=50, semi_len0=100, semi_len1=130, intensity=2)
        assert (test.signal.data == 2).all()

    def test_with_downscale_blur_default(self):
        test = mdtd.MakeTestData(size_x=120, size_y=100)
        test.add_disk_ellipse(x0=50, y0=50, semi_len0=10, semi_len1=13, intensity=2)

    def test_disk_cover_nothing(self):
        test = mdtd.MakeTestData(
            size_x=120, size_y=100, default=False, blur=False, downscale=False
        )
        test.add_disk_ellipse(x0=350, y0=50, semi_len0=10, semi_len1=10, intensity=2)
        assert not test.signal.data.any()

    def test_disk_semi_len0_very_small(self):
        test = mdtd.MakeTestData(
            size_x=120, size_y=100, default=False, blur=False, downscale=False
        )
        x0, y0 = 50, 40
        intensity = 2
        test.add_disk_ellipse(
            x0=x0, y0=y0, semi_len0=0.1, semi_len1=1000, intensity=intensity, rotation=0
        )
        data = test.signal.data
        assert (data[:, x0] == intensity).all()
        data[:, x0] = 0
        assert not data.any()

    def test_disk_semi_len1_very_small(self):
        test = mdtd.MakeTestData(
            size_x=120, size_y=100, default=False, blur=False, downscale=False
        )
        x0, y0 = 50, 40
        intensity = 2
        test.add_disk_ellipse(
            x0=x0, y0=y0, semi_len0=1000, semi_len1=0.1, intensity=intensity, rotation=0
        )
        data = test.signal.data
        assert (data[y0, :] == intensity).all()
        data[y0, :] = 0
        assert not data.any()

    def test_disk_semi_len1_very_small_rot_90(self):
        test = mdtd.MakeTestData(
            size_x=120, size_y=100, default=False, blur=False, downscale=False
        )
        x0, y0 = 50, 40
        intensity = 2
        test.add_disk_ellipse(
            x0=x0,
            y0=y0,
            semi_len0=1000,
            semi_len1=0.1,
            intensity=intensity,
            rotation=np.pi / 2,
        )
        data = test.signal.data
        assert (data[:, x0] == intensity).all()
        data[:, x0] = 0
        assert not data.any()

    def test_disk_semi_len1_very_small_180(self):
        test = mdtd.MakeTestData(
            size_x=120, size_y=100, default=False, blur=False, downscale=False
        )
        x0, y0 = 50, 40
        intensity = 2
        test.add_disk_ellipse(
            x0=x0,
            y0=y0,
            semi_len0=1000,
            semi_len1=0.1,
            intensity=intensity,
            rotation=np.pi,
        )
        data = test.signal.data
        assert (data[y0, :] == intensity).all()
        data[y0, :] = 0
        assert not data.any()

    def test_repr(self):
        test = mdtd.MakeTestData()
        test.add_disk_ellipse(x0=70)
        repr_string = test.z_list[0].__repr__()
        assert str(70) in repr_string


class TestMakeDiffractionTestDataDisksRing:
    def test_downscale(self):
        test0 = mdtd.MakeTestData(
            size_x=120, size_y=100, downscale=True, blur=False, default=False
        )
        test0.add_ring_ellipse(x0=50, y0=50, semi_len0=8, semi_len1=13, intensity=2)
        test1 = mdtd.MakeTestData(
            size_x=120, size_y=100, downscale=False, blur=False, default=False
        )
        test1.add_ring_ellipse(x0=50, y0=50, semi_len0=8, semi_len1=13, intensity=2)
        assert not (test0.signal.data == test1.signal.data).all()

    def test_blur(self):
        test0 = mdtd.MakeTestData(
            size_x=120, size_y=100, downscale=False, blur=True, default=False
        )
        test0.add_ring_ellipse(x0=50, y0=50, semi_len0=8, semi_len1=13, intensity=2)
        test1 = mdtd.MakeTestData(
            size_x=120, size_y=100, downscale=False, blur=False, default=False
        )
        test1.add_ring_ellipse(x0=50, y0=50, semi_len0=8, semi_len1=13, intensity=2)
        assert not (test0.signal.data == test1.signal.data).all()

    def test_very_large_radius_no_cover(self):
        test = mdtd.MakeTestData(size_x=120, size_y=100)
        test.add_ring_ellipse(x0=50, y0=50, semi_len0=200, semi_len1=159, intensity=2)
        assert not test.signal.data.any()

    def test_lw_r(self):
        test0 = mdtd.MakeTestData(size_x=120, size_y=100)
        test0.add_ring_ellipse(x0=50, y0=50, semi_len0=20, semi_len1=30)
        test1 = mdtd.MakeTestData(size_x=120, size_y=100)
        test1.add_ring_ellipse(x0=50, y0=50, semi_len0=20, semi_len1=30, lw_r=3)
        assert test0.signal.data.sum() < test1.signal.data.sum()

    def test_rotation(self):
        test0 = mdtd.MakeTestData(size_x=120, size_y=100)
        test0.add_ring_ellipse(x0=50, y0=50, semi_len0=20, semi_len1=30)
        test1 = mdtd.MakeTestData(size_x=120, size_y=100)
        test1.add_ring_ellipse(x0=50, y0=50, semi_len0=20, semi_len1=30, rotation=1)
        assert not np.all(np.equal(test0.signal.data, test1.signal.data))

    def test_repr(self):
        test = mdtd.MakeTestData()
        test.add_ring_ellipse(x0=70)
        repr_string = test.z_list[0].__repr__()
        assert str(70) in repr_string


class TestGenerate4dData:
    def test_simple0(self):
        mdtd.generate_4d_data()

    def test_all_arguments(self):
        s = mdtd.generate_4d_data(
            probe_size_x=10,
            probe_size_y=10,
            image_size_x=50,
            image_size_y=50,
            disk_x=20,
            disk_y=20,
            disk_r=5,
            disk_I=30,
            ring_x=None,
            ring_e_x=None,
            blur=True,
            blur_sigma=1,
            downscale=True,
            add_noise=True,
            noise_amplitude=2,
        )
        assert s.axes_manager.shape == (10, 10, 50, 50)

    def test_different_size(self):
        s = mdtd.generate_4d_data(
            probe_size_x=5,
            probe_size_y=7,
            ring_x=None,
            ring_e_x=None,
            image_size_x=30,
            image_size_y=50,
        )
        ax = s.axes_manager
        assert ax.navigation_dimension == 2
        assert ax.signal_dimension == 2
        assert ax.navigation_shape == (5, 7)
        assert ax.signal_shape == (30, 50)

    def test_disk_outside_image(self):
        s = mdtd.generate_4d_data(
            probe_size_x=6,
            probe_size_y=4,
            image_size_x=40,
            image_size_y=40,
            ring_x=None,
            ring_e_x=None,
            disk_x=1000,
            disk_y=1000,
            disk_r=5,
        )
        assert (s.data == 0).all()

    def test_disk_cover_whole_image(self):
        s = mdtd.generate_4d_data(
            probe_size_x=6,
            probe_size_y=4,
            image_size_x=20,
            image_size_y=20,
            ring_x=None,
            ring_e_x=None,
            disk_x=10,
            disk_y=10,
            disk_r=40,
            disk_I=50,
            blur=False,
            downscale=False,
        )
        assert (s.data == 50).all()

    def test_disk_position_array(self):
        ps_x, ps_y, intensity = 4, 7, 30
        disk_x = np.random.randint(5, 35, size=(ps_y, ps_x))
        disk_y = np.random.randint(5, 45, size=(ps_y, ps_x))
        s = mdtd.generate_4d_data(
            probe_size_x=ps_x,
            probe_size_y=ps_y,
            image_size_x=40,
            image_size_y=50,
            ring_x=None,
            ring_e_x=None,
            disk_x=disk_x,
            disk_y=disk_y,
            disk_r=1,
            disk_I=intensity,
            blur=False,
            downscale=False,
        )
        for x in range(ps_x):
            for y in range(ps_y):
                cX, cY = disk_x[y, x], disk_y[y, x]
                sl = np.s_[cY - 1 : cY + 2, cX - 1 : cX + 2]
                im = s.inav[x, y].data[:]
                assert (im[sl] == intensity).all()
                im[sl] = 0
                assert not im.any()

    def test_disk_ring_outside_image(self):
        s = mdtd.generate_4d_data(
            probe_size_x=6,
            probe_size_y=4,
            image_size_x=40,
            image_size_y=40,
            disk_x=1000,
            disk_y=1000,
            disk_r=5,
            ring_x=1000,
            ring_y=1000,
            ring_r=10,
            ring_e_x=None,
        )
        assert (s.data == 0).all()

    def test_ring_center(self):
        im_x, im_y, x, y = 120, 100, 40, 51
        s = mdtd.generate_4d_data(
            probe_size_x=4,
            probe_size_y=5,
            image_size_x=im_x,
            image_size_y=im_y,
            disk_x=x,
            disk_y=y,
            disk_r=10,
            disk_I=0,
            ring_x=x,
            ring_y=y,
            ring_r=30,
            ring_I=5,
            ring_e_x=None,
            blur=False,
            downscale=False,
        )
        s_bs = s.get_direct_beam_position(method="center_of_mass")
        x_new = (im_x / 2) - x
        y_new = (im_y / 2) - y
        assert (s_bs.isig[0].data == x_new).all()
        assert (s_bs.isig[1].data == y_new).all()

    def test_ring_ellipse_center(self):
        im_x, im_y, x, y = 120, 100, 40, 51
        s = mdtd.generate_4d_data(
            probe_size_x=4,
            probe_size_y=5,
            image_size_x=120,
            image_size_y=100,
            disk_x=None,
            ring_x=None,
            ring_e_x=x,
            ring_e_y=y,
            blur=False,
            downscale=False,
        )
        s_bs = s.get_direct_beam_position(method="center_of_mass")
        x_new = (im_x / 2) - x
        y_new = (im_y / 2) - y
        assert (s_bs.isig[0].data == x_new).all()
        assert (s_bs.isig[1].data == y_new).all()

    def test_input_numpy_array(self):
        size = (20, 10)
        disk_x = np.random.randint(5, 35, size=size)
        disk_y = np.random.randint(5, 45, size=size)
        disk_r = np.random.randint(5, 9, size=size)
        disk_I = np.random.randint(50, 100, size=size)
        ring_x = np.random.randint(5, 35, size=size)
        ring_y = np.random.randint(5, 45, size=size)
        ring_r = np.random.randint(10, 15, size=size)
        ring_I = np.random.randint(1, 30, size=size)
        ring_lw = np.random.randint(1, 5, size=size)
        ring_e_x = np.random.randint(20, 30, size)
        ring_e_y = np.random.randint(20, 30, size)
        ring_e_semi_len0 = np.random.randint(10, 20, size)
        ring_e_semi_len1 = np.random.randint(10, 20, size)
        ring_e_r = np.random.random(size) * np.pi
        ring_e_lw = np.random.randint(1, 3, size)

        mdtd.generate_4d_data(
            probe_size_x=10,
            probe_size_y=20,
            image_size_x=40,
            image_size_y=50,
            disk_x=disk_x,
            disk_y=disk_y,
            disk_I=disk_I,
            disk_r=disk_r,
            ring_x=ring_x,
            ring_y=ring_y,
            ring_r=ring_r,
            ring_I=ring_I,
            ring_lw=ring_lw,
            ring_e_x=ring_e_x,
            ring_e_y=ring_e_y,
            ring_e_semi_len0=ring_e_semi_len0,
            ring_e_semi_len1=ring_e_semi_len1,
            ring_e_r=ring_e_r,
            ring_e_lw=ring_e_lw,
        )


class TestGetEllipticalRing:
    def test_simple(self):
        s = Signal2D(np.zeros((110, 200)))
        x, y, semi_len0, semi_len1, rotation = 60, 70, 12, 9, 0.2
        xx, yy = np.meshgrid(s.axes_manager[0].axis, s.axes_manager[1].axis)
        ellipse_image = mdtd._get_elliptical_ring(
            xx, yy, x, y, semi_len0, semi_len1, rotation, 3
        )
        assert ellipse_image.any()

    def test_cover_no_signal1(self):
        s = Signal2D(np.zeros((50, 56)))
        x, y, semi_len0, semi_len1, rotation = 523, 620, 20, 30, 2.0
        xx, yy = np.meshgrid(s.axes_manager[0].axis, s.axes_manager[1].axis)
        ellipse_image = mdtd._get_elliptical_ring(
            xx, yy, x, y, semi_len0, semi_len1, rotation
        )
        assert not ellipse_image.any()

    def test_cover_no_signal2(self):
        s = Signal2D(np.zeros((50, 56)))
        x, y, semi_len0, semi_len1, rotation = 20, 20, 90, 130, 2.0
        xx, yy = np.meshgrid(s.axes_manager[0].axis, s.axes_manager[1].axis)
        ellipse_image = mdtd._get_elliptical_ring(
            xx, yy, x, y, semi_len0, semi_len1, rotation
        )
        assert not ellipse_image.any()


class TestGetEllipticalMask:
    def test_simple(self):
        s = Signal2D(np.zeros((110, 200)))
        xx, yy = np.meshgrid(s.axes_manager[0].axis, s.axes_manager[1].axis)
        x, y, semi_len0, semi_len1, rotation = 60, 70, 12, 9, 0.2
        ellipse_image = mdtd._get_elliptical_disk(
            xx, yy, x, y, semi_len0, semi_len1, rotation
        )
        assert s.data.shape == ellipse_image.shape

    def test_cover_all_signal(self):
        s = Signal2D(np.zeros((50, 56)))
        xx, yy = np.meshgrid(s.axes_manager[0].axis, s.axes_manager[1].axis)
        x, y, semi_len0, semi_len1, rotation = 23, 20, 200, 200, 0.0
        ellipse_image = mdtd._get_elliptical_disk(
            xx, yy, x, y, semi_len0, semi_len1, rotation
        )
        assert s.data.shape == ellipse_image.shape
        assert ellipse_image.all()

    def test_cover_no_signal(self):
        s = Signal2D(np.zeros((50, 56)))
        xx, yy = np.meshgrid(s.axes_manager[0].axis, s.axes_manager[1].axis)
        x, y, semi_len0, semi_len1, rotation = 523, 620, 20, 30, 2.0
        ellipse_image = mdtd._get_elliptical_disk(
            xx, yy, x, y, semi_len0, semi_len1, rotation
        )
        assert not ellipse_image.any()


class TestMake4dPeakArrayTestData:
    def test_simple(self):
        xf, yf = np.ones((2, 3)), np.ones((2, 3))
        semi0, semi1, rot = np.ones((2, 3)), np.ones((2, 3)), np.ones((2, 3))
        peak_array = mdtd._make_4d_peak_array_test_data(xf, yf, semi0, semi1, rot)
        assert peak_array.shape == xf.shape

    def test_xf_yf(self):
        xf = np.random.randint(10, 20, size=(2, 3))
        yf = np.random.randint(39, 60, size=(2, 3))
        semi0, semi1, rot = np.ones((2, 3)), np.ones((2, 3)), np.ones((2, 3))
        peak_array = mdtd._make_4d_peak_array_test_data(
            xf, yf, semi0, semi1, rot, nt=1000
        )
        for iy, ix in np.ndindex(xf.shape):
            x_pos_mean = peak_array[iy, ix][:, 1].mean()
            assert approx(x_pos_mean, abs=0.01) == xf[iy, ix]
            y_pos_mean = peak_array[iy, ix][:, 0].mean()
            assert approx(y_pos_mean, abs=0.01) == yf[iy, ix]

    def test_semi_lengths(self):
        semi0 = np.random.randint(10, 20, size=(2, 3))
        semi1 = np.random.randint(10, 20, size=(2, 3))
        xf, yf, rot = np.ones((2, 3)) * 10, np.ones((2, 3)), np.zeros((2, 3))
        peak_array0 = mdtd._make_4d_peak_array_test_data(
            xf, yf, semi0, semi0, rot, nt=1000
        )
        peak_array1 = mdtd._make_4d_peak_array_test_data(
            xf, yf, semi1, semi1, rot, nt=1000
        )
        for iy, ix in np.ndindex(xf.shape):
            semi0_max = semi0[iy, ix] + xf[iy, ix]
            assert approx(peak_array0[iy, ix][:, 1].max()) == semi0_max
            semi1_max = semi1[iy, ix] + yf[iy, ix]
            assert approx(peak_array1[iy, ix][:, 0].max()) == semi1_max

    def test_rotation(self):
        rot0, rot1 = np.zeros((2, 3)), np.ones((2, 3)) * np.pi / 2
        xf, yf = np.ones((2, 3)), np.ones((2, 3))
        semi0, semi1 = np.ones((2, 3)) * 5, np.ones((2, 3))
        pa0 = mdtd._make_4d_peak_array_test_data(xf, yf, semi0, semi1, rot0, nt=1000)
        pa1 = mdtd._make_4d_peak_array_test_data(xf, yf, semi0, semi1, rot1, nt=1000)
        for i in np.ndindex(pa1.shape):
            assert not np.all(np.array_equal(pa0[i], pa1[i]))

    def test_nt(self):
        p = np.ones((2, 3))
        nt0, nt1 = 100, 200
        pa0 = mdtd._make_4d_peak_array_test_data(p, p, p, p, p, nt=nt0)
        pa1 = mdtd._make_4d_peak_array_test_data(p, p, p, p, p, nt=nt1)
        assert len(pa0[0, 0][:, 0]) == nt0
        assert len(pa1[0, 0][:, 0]) == nt1


class TestDiffractionTestImage:
    def test_simple(self):
        mdtd.DiffractionTestImage()

    def test_repr(self):
        di = mdtd.DiffractionTestImage(image_y=111)
        repr_string = di.__repr__()
        assert str(111) in repr_string

    def test_plot(self):
        di = mdtd.DiffractionTestImage()
        di.plot()

    def test_get_signal(self):
        di = mdtd.DiffractionTestImage()
        s = di.get_signal()
        assert hasattr(s, "plot")

    def test_copy(self):
        di0 = mdtd.DiffractionTestImage(intensity_noise=False)
        di0.add_cubic_disks(20, 20)
        di1 = di0.copy()
        data0 = di0.get_diffraction_test_image()
        data1 = di1.get_diffraction_test_image()
        assert (data0 == data1).all()

    def test_get_diffraction_test_image(self):
        di = mdtd.DiffractionTestImage()
        data = di.get_diffraction_test_image()
        assert hasattr(data, "__array__")

    def test_get_diffraction_test_image_dtype(self):
        di = mdtd.DiffractionTestImage()
        data = di.get_diffraction_test_image(dtype=np.float32)
        assert data.dtype == np.float32
        data = di.get_diffraction_test_image(dtype=np.float64)
        assert data.dtype == np.float64

    def test_add_disk(self):
        di = mdtd.DiffractionTestImage(
            diff_intensity_reduction=False, intensity_noise=False, blur=0
        )
        di.add_disk(10, 40, intensity=3.0)
        di.add_disk(128, 128, intensity=4.0)
        data = di.get_diffraction_test_image()
        assert data[40, 10] == 3.0
        assert data[128, 128] == 4.0

    def test_add_cubic_disks_n(self):
        di = mdtd.DiffractionTestImage(
            diff_intensity_reduction=False, intensity_noise=False, blur=0
        )
        di0 = di.copy()
        di1 = di.copy()
        di0.add_cubic_disks(20, 20, n=1)
        di1.add_cubic_disks(20, 20, n=2)
        data0 = di0.get_diffraction_test_image()
        data1 = di1.get_diffraction_test_image()
        assert data0.sum() < data1.sum()

    def test_add_cubic_disks_intensity(self):
        di = mdtd.DiffractionTestImage(
            diff_intensity_reduction=False,
            intensity_noise=False,
            blur=0,
            image_x=100,
            image_y=100,
        )
        di = di.copy()
        di.add_cubic_disks(20, 20, intensity=25.0, n=1)
        data = di.get_diffraction_test_image()
        assert data[30, 30] == 25.0

    def test_add_background_lorentz(self):
        di = mdtd.DiffractionTestImage(
            diff_intensity_reduction=False, intensity_noise=False, blur=0
        )
        data0 = di.get_diffraction_test_image()
        di.add_background_lorentz(intensity=10.0, width=5.0)
        data1 = di.get_diffraction_test_image()
        assert data0[50, 50] == 0.0
        assert data1[50, 50] != 0.0
        di.add_background_lorentz(intensity=10.0, width=20.0)
        data2 = di.get_diffraction_test_image()
        assert data2[50, 50] != data1[50, 50]
        di.add_background_lorentz(intensity=90.0, width=20.0)
        data3 = di.get_diffraction_test_image()
        assert data3[50, 50] > data2[50, 50]

    def test_image_x_y(self):
        x, y = 100, 100
        di = mdtd.DiffractionTestImage(image_x=x, image_y=y)
        data = di.get_diffraction_test_image()
        assert data.shape == (100, 100)
        s = di.get_signal()
        assert s.axes_manager.signal_shape == (100, 100)

    def test_disk_r(self):
        di = mdtd.DiffractionTestImage(
            image_x=50, image_y=50, disk_r=10, intensity_noise=False
        )
        di.add_disk(25, 25, intensity=2.0)
        data = di.get_diffraction_test_image()
        assert data[0, 0] == 0
        assert data[-1, 0] == 0
        assert data[0, -1] == 0
        assert data[-1, -1] == 0
        di.disk_r = 40
        data = di.get_diffraction_test_image()
        assert data[0, 0] == 2.0

    def test_disk_bad_input(self):
        di = mdtd.DiffractionTestImage()
        with pytest.raises(ValueError):
            di.add_disk(1.5, 5)
        with pytest.raises(ValueError):
            di.add_disk(5, 1.5)

    def test_rotation(self):
        di = mdtd.DiffractionTestImage(
            diff_intensity_reduction=False, intensity_noise=False, blur=0
        )
        di.add_disk(10, 10, intensity=2.0)
        data0 = di.get_diffraction_test_image()
        di.rotation = 180
        data1 = di.get_diffraction_test_image()
        assert data0[10, 10] == 2.0
        assert data1[-10, -10] == 2.0

    def test_blur(self):
        di = mdtd.DiffractionTestImage(
            diff_intensity_reduction=False, intensity_noise=False, blur=0
        )
        di.add_disk(10, 10, intensity=2.0)
        data0 = di.get_diffraction_test_image()
        di.blur = 3.0
        data1 = di.get_diffraction_test_image()
        assert data0[10, 10] == 2.0
        assert data1[10, 10] != 2.0

    def test_diff_intensity_reduction(self):
        di = mdtd.DiffractionTestImage(
            diff_intensity_reduction=False, intensity_noise=False, blur=0
        )
        di.add_disk(10, 10, intensity=2.0)
        data0 = di.get_diffraction_test_image()
        di.diff_intensity_reduction = 1.0
        data1 = di.get_diffraction_test_image()
        assert data0[10, 10] == 2.0
        assert data1[10, 10] != 2.0


class TestDiffractionTestDataset:
    def test_simple(self):
        mdtd.DiffractionTestDataset()

    def test_repr(self):
        dtd = mdtd.DiffractionTestDataset(probe_x=21)
        repr_string = dtd.__repr__()
        assert str(21) in repr_string

    def test_dataset_size(self):
        dtd = mdtd.DiffractionTestDataset(
            probe_x=5, probe_y=15, detector_x=101, detector_y=200
        )
        assert dtd.data.shape == (5, 15, 101, 200)

    def test_plot(self):
        dtd = mdtd.DiffractionTestDataset()
        dtd.plot()

    def test_get_signal(self):
        dtd = mdtd.DiffractionTestDataset()
        s = dtd.get_signal()
        assert hasattr(s, "plot")

    def test_add_diffraction_image(self):
        dtd = mdtd.DiffractionTestDataset(noise=False)
        di = mdtd.DiffractionTestImage(intensity_noise=False)
        di.add_cubic_disks(20, 20, n=1)
        data = di.get_diffraction_test_image()
        dtd.add_diffraction_image(di)
        for ix, iy in np.ndindex(dtd.data.shape[:2]):
            assert (dtd.data[ix, iy] == data).all()



# -*- coding: utf-8 -*-
# Copyright 2016-2025 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np

from hyperspy.signals import Signal2D
from hyperspy.roi import Line2DROI

from pyxem.signals import ElectronDiffraction2D
from pyxem.libraries import CalibrationDataLibrary


@pytest.fixture
def library(diffraction_pattern):
    dp = diffraction_pattern.mean((0, 1))
    im = Signal2D(np.ones((10, 10)))
    cdl = CalibrationDataLibrary(
        au_x_grating_dp=dp, au_x_grating_im=im, moo3_dp=dp, moo3_im=im
    )
    return cdl


def test_initialization_dtype(library):
    assert isinstance(library.au_x_grating_dp, ElectronDiffraction2D)


class TestPlotData:
    def test_plot_au_x_grating_dp(self, library):
        library.plot_calibration_data(data_to_plot="au_x_grating_dp")

    def test_plot_au_x_grating_im(self, library):
        library.plot_calibration_data(data_to_plot="au_x_grating_im")

    def test_plot_moo3_dp(self, library):
        library.plot_calibration_data(data_to_plot="moo3_dp")

    def test_plot_moo3_im(self, library):
        library.plot_calibration_data(data_to_plot="moo3_im")

    def test_plot_invalid(self, library):
        with pytest.raises(ValueError, match="Please specify valid data_to_plot"):
            library.plot_calibration_data(data_to_plot="no_data")

    def test_plot_au_x_grating_dp_with_roi(self, library):
        line = Line2DROI(x1=1, y1=1, x2=3, y2=3, linewidth=1.0)
        library.plot_calibration_data(data_to_plot="au_x_grating_dp", roi=line)
